# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2022/12/31 16:03
describe: QMT 量化交易平台接口

需要额外安装：pyautogui，xtquant

xtquant 的下载地址：http://dict.thinktrader.net/nativeApi/download_xtquant.html?id=7zqjlm
"""
import os
import time
import random
import czsc
import loguru
import pyautogui
import subprocess
import pandas as pd
from typing import List
from tqdm import tqdm

# from loguru import logger
from datetime import datetime, timedelta
from czsc.objects import Freq, RawBar
from czsc.fsa.im import IM
from czsc.traders.base import CzscTrader
from czsc.utils import resample_bars
from xtquant import xtconstant
from xtquant import xtdata
from xtquant.xttrader import XtQuantTrader, XtQuantTraderCallback
from xtquant.xttype import StockAccount

dt_fmt = "%Y-%m-%d %H:%M:%S"


def find_exe_window(title, **kwargs):
    """windows系统：根据 title 查找 window"""
    logger = kwargs.get("logger", loguru.logger)
    windows = pyautogui.getWindowsWithTitle(title)
    if len(windows) > 1:
        logger.warning(f"找到多个 {title} 窗口，数量：{len(windows)}；请检查是否有多个程序实例")
        return None

    if len(windows) == 0:
        return None

    return windows[0]


def close_exe_window(title, **kwargs):
    """windows系统：关闭 exe 应用程序

    :param title: 程序标题，支持模糊匹配，不需要完全匹配

        查看应用的标题（Title）通常指的是获取正在运行的应用程序窗口的标题栏上的文字。
        在Windows操作系统中，你可以通过几种不同的方法来查看应用的标题：

        1. 使用任务管理器
            1. 按下 Ctrl + Shift + Esc 打开任务管理器。
            2. 在“进程”标签页中，找到你想要查看标题的应用程序。
            3. 在“应用程序”列表中，找到对应的应用程序，其标题通常显示在“名称”列中。

        2. 使用Alt+Tab快捷键
            1. 按下 Alt + Tab 快捷键，可以快速切换到正在运行的应用程序。
            2. 在弹出的窗口中，你可以看到每个应用程序的缩略图和标题。

        3. 使用任务栏
            1. 将鼠标悬停在任务栏上的应用程序图标上。
            2. 通常，任务栏会显示该应用程序的标题。
    """
    logger = kwargs.get("logger", loguru.logger)

    window = find_exe_window(title)
    if window:
        window.activate()
        window.close()
        pyautogui.press("enter")
    else:
        logger.error(f"没有找到 {title} 程序")


def wait_qmt_ready(timeout=60, **kwargs):
    """等待 QMT 窗口就绪"""
    logger = kwargs.get("logger", loguru.logger)

    start = time.time()
    while time.time() - start < timeout:
        x = xtdata.get_full_tick(code_list=["000001.SZ"])
        if x and x["000001.SZ"]["lastPrice"] > 0:
            return True

        logger.warning("等待 QMT 窗口就绪")
        time.sleep(1)
    return False


def initialize_qmt(**kwargs):
    """初始化 QMT 交易端

    :param kwargs:
        mini_qmt_dir: str, mini qmt 目录
        account_id: str, 账户id
        callback_params: dict, TraderCallback 回调类的参数
    :return: xtt, acc
        xtt - XtQuantTrader
        acc - StockAccount
    """
    logger = kwargs.get("logger", loguru.logger)

    import random

    mini_qmt_dir = kwargs.get("mini_qmt_dir")
    account_id = kwargs.get("account_id")

    session = random.randint(10000, 20000)
    callback = TraderCallback(logger=logger, **kwargs.get("callback_params", {}))
    xtt = XtQuantTrader(mini_qmt_dir, session=session, callback=callback)
    acc = StockAccount(account_id, "STOCK")
    xtt.start()
    xtt.connect()
    assert xtt.connected, "交易服务器连接失败"
    _res = xtt.subscribe(acc)
    assert _res == 0, "账号订阅失败"
    return xtt, acc


def start_qmt_exe(acc, pwd, qmt_exe, title, max_retry=6, **kwargs):
    """windows系统：启动 QMT，并登录

    :param acc: QMT 账号
    :param pwd: QMT 密码
    :param qmt_exe: QMT 应用路径，如 D:\\国金证券QMT交易端\\bin.x64\\XtItClient.exe
    :param title: QMT 窗口标题，如 国金证券QMT交易端
    :param max_retry: 最大重试次数
    """
    logger = kwargs.get("logger", loguru.logger)

    wait_seconds = kwargs.get("wait_seconds", 6)
    i = 0
    while not find_exe_window(title):
        if i > max_retry:
            logger.warning(f"QMT连续{i}次尝试依旧无法启动，请人工检查!")
            break

        close_exe_window(title)

        i += 1
        try:
            subprocess.Popen(qmt_exe)
            time.sleep(wait_seconds)

            qmt_windows = find_exe_window(title)
            if not qmt_windows:
                continue

            qmt_windows.activate()
            pyautogui.typewrite(acc)
            pyautogui.press("tab")
            pyautogui.typewrite(pwd)
            pyautogui.press("enter")
            qmt_windows.activate()
            time.sleep(wait_seconds)

        except Exception as e:
            logger.exception(f"启动QMT失败：{e}")

    logger.info(f"启动 QMT 成功，账号：{acc}")


def format_stock_kline(kline: pd.DataFrame, freq: Freq) -> List[RawBar]:
    """QMT A股市场K线数据转换

    :param kline: QMT 数据接口返回的K线数据
                         time   open   high    low  close  volume      amount  \
        0 2022-12-01 10:15:00  13.22  13.22  13.16  13.18   20053  26432861.0
        1 2022-12-01 10:20:00  13.18  13.19  13.15  13.15   32667  43002512.0
        2 2022-12-01 10:25:00  13.16  13.18  13.13  13.16   32466  42708049.0
        3 2022-12-01 10:30:00  13.16  13.19  13.13  13.18   15606  20540461.0
        4 2022-12-01 10:35:00  13.20  13.25  13.19  13.22   29959  39626170.0
              symbol
        0  000001.SZ
        1  000001.SZ
        2  000001.SZ
        3  000001.SZ
        4  000001.SZ
    :param freq: K线周期
    :return: 转换好的K线数据
    """
    bars = []
    dt_key = "time"
    kline = kline.sort_values(dt_key, ascending=True, ignore_index=True)
    records = kline.to_dict("records")

    for i, record in enumerate(records):
        # 将每一根K线转换成 RawBar 对象
        bar = RawBar(
            symbol=record["symbol"],
            dt=pd.to_datetime(record[dt_key]),
            id=i,
            freq=freq,
            open=record["open"],
            close=record["close"],
            high=record["high"],
            low=record["low"],
            vol=record["volume"] * 100 if record["volume"] else 0,  # 成交量，单位：股
            amount=record["amount"] if record["amount"] > 0 else 0,  # 成交额，单位：元
        )
        bars.append(bar)
    return bars


def get_kline(symbol, period, start_time, end_time, count=-1, dividend_type="front_ratio", **kwargs):
    """获取 QMT K线数据，实盘、回测通用

    :param symbol: 股票代码 例如：300001.SZ
    :param period: 周期 分笔"tick" 分钟线"1m"/"5m" 日线"1d"
    :param start_time: 开始时间
    :param end_time: 结束时间
    :param count: 数量 -1全部，n: 从结束时间向前数n个
    :param dividend_type: 除权类型"none" "front" "back" "front_ratio" "back_ratio"

    :return: df Dataframe格式的数据，样例如下
                         time   open   high    low  close  volume      amount  \
        0 2022-12-01 10:15:00  13.22  13.22  13.16  13.18   20053  26432861.0
        1 2022-12-01 10:20:00  13.18  13.19  13.15  13.15   32667  43002512.0
        2 2022-12-01 10:25:00  13.16  13.18  13.13  13.16   32466  42708049.0
        3 2022-12-01 10:30:00  13.16  13.19  13.13  13.18   15606  20540461.0
        4 2022-12-01 10:35:00  13.20  13.25  13.19  13.22   29959  39626170.0
              symbol
        0  000001.SZ
        1  000001.SZ
        2  000001.SZ
        3  000001.SZ
        4  000001.SZ
    """
    logger = kwargs.get("logger", loguru.logger)

    start_time = pd.to_datetime(start_time).strftime("%Y%m%d%H%M%S")
    if "1d" == period:
        end_time = pd.to_datetime(end_time).replace(hour=15, minute=0).strftime("%Y%m%d%H%M%S")
    else:
        end_time = pd.to_datetime(end_time).strftime("%Y%m%d%H%M%S")

    if kwargs.get("download_hist", True):
        xtdata.download_history_data(symbol, period=period, start_time=start_time, end_time=end_time)

    field_list = ["time", "open", "high", "low", "close", "volume", "amount"]
    data = xtdata.get_market_data(
        field_list,
        stock_list=[symbol],
        period=period,
        count=count,
        dividend_type=dividend_type,
        start_time=start_time,
        end_time=end_time,
        fill_data=kwargs.get("fill_data", False),
    )

    df = pd.DataFrame({key: value.values[0] for key, value in data.items()})
    df["time"] = pd.to_datetime(df["time"], unit="ms") + pd.to_timedelta("8H")
    df.reset_index(inplace=True, drop=True)
    df["symbol"] = symbol
    df = df.dropna()

    if kwargs.get("df", True):
        return df
    else:
        freq_map = {"1m": Freq.F1, "5m": Freq.F5, "1d": Freq.D}
        return format_stock_kline(df, freq=freq_map[period])


def get_raw_bars(symbol, freq, sdt, edt, fq="前复权", **kwargs) -> List[RawBar]:
    """获取 CZSC 库定义的标准 RawBar 对象列表

    :param symbol: 标的代码
    :param freq: 周期
    :param sdt: 开始时间
    :param edt: 结束时间
    :param fq: 除权类型
    :param kwargs:
    :return:
    """
    logger = kwargs.get("logger", loguru.logger)

    freq = Freq(freq)
    if freq == Freq.F1:
        period = "1m"
    elif freq in [Freq.F5, Freq.F15, Freq.F30, Freq.F60]:
        period = "5m"
    else:
        period = "1d"

    if fq == "前复权":
        dividend_type = "front_ratio"
    elif fq == "后复权":
        dividend_type = "back_ratio"
    else:
        assert fq == "不复权"
        dividend_type = "none"

    kline = get_kline(
        symbol, period, sdt, edt, dividend_type=dividend_type, download_hist=kwargs.get("download_hist", True), df=True
    )
    if kline.empty:
        return []

    kline["dt"] = pd.to_datetime(kline["time"])
    kline["vol"] = kline["volume"]
    bars = resample_bars(kline, freq, raw_bars=True)
    return bars


def get_symbols(step):
    """获取择时策略投研不同阶段对应的标的列表

    :param step: 投研阶段
    :return: 标的列表
    """
    stocks = xtdata.get_stock_list_in_sector("沪深A股")
    stocks_map = {
        "index": [
            "000905.SH",
            "000016.SH",
            "000300.SH",
            "000001.SH",
            "000852.SH",
            "399001.SZ",
            "399006.SZ",
            "399376.SZ",
            "399377.SZ",
            "399317.SZ",
            "399303.SZ",
        ],
        "stock": stocks,
        "check": ["000001.SZ"],
        "train": stocks[:200],
        "valid": stocks[200:600],
        "etfs": [
            "512880.SH",
            "518880.SH",
            "515880.SH",
            "513050.SH",
            "512690.SH",
            "512660.SH",
            "512400.SH",
            "512010.SH",
            "512000.SH",
            "510900.SH",
            "510300.SH",
            "510500.SH",
            "510050.SH",
            "159992.SZ",
            "159985.SZ",
            "159981.SZ",
            "159949.SZ",
            "159915.SZ",
        ],
    }
    if step.upper() == "ALL":
        return stocks_map["index"] + stocks_map["stock"] + stocks_map["etfs"]

    return stocks_map[step]


def is_trade_time(dt: datetime = datetime.now()):
    """判断指定时间是否是交易时间"""
    hm = dt.strftime("%H:%M")
    if hm < "09:15" or hm > "15:00":
        return False
    else:
        return True


def is_stock_trade_time(dt: datetime = datetime.now()):
    """判断指定时间是否是股票交易时间"""
    hm = dt.strftime("%H:%M")
    if "09:30" <= hm < "11:30":
        return True
    if "13:00" <= hm < "15:00":
        return True
    return False


def is_trade_day(dt: datetime = datetime.now()):
    """判断指定日期是否是交易日"""
    date = dt.strftime("%Y%m%d")
    return True if xtdata.get_trading_dates("SH", date, date) else False


class TraderCallback(XtQuantTraderCallback):
    """基础回调类，主要是一些日志和IM通知功能"""

    def __init__(self, **kwargs):
        super(TraderCallback, self).__init__()
        self.kwargs = kwargs
        self.logger = kwargs.get("logger", loguru.logger)

        if kwargs.get("feishu_app_id", None) and kwargs.get("feishu_app_secret", None):
            self.im = IM(app_id=kwargs["feishu_app_id"], app_secret=kwargs["feishu_app_secret"])
            self.members = kwargs["feishu_members"]
        else:
            self.im = None
            self.members = None

        # 推送模式：detail-详细模式，summary-汇总模式
        self.feishu_push_mode = kwargs.get("feishu_push_mode", "detail")

        file_log = kwargs.get("file_log", None)
        if file_log:
            self.logger.add(file_log, rotation="1 day", encoding="utf-8", enqueue=True)
        self.file_log = file_log
        self.logger.info(f"TraderCallback init: {kwargs}")

    def push_message(self, msg: str, msg_type="text"):
        """批量推送消息"""
        if self.im and self.members:
            for member in self.members:
                try:
                    if msg_type == "text":
                        self.im.send_text(msg, member)
                    elif msg_type == "image":
                        self.im.send_image(msg, member)
                    elif msg_type == "file":
                        self.im.send_file(msg, member)
                    else:
                        self.logger.error(f"不支持的消息类型：{msg_type}")
                except Exception as e:
                    self.logger.error(f"推送消息失败：{e}")

    def on_disconnected(self):
        """连接断开"""
        self.logger.info("connection lost")
        if is_trade_time():
            self.push_message("连接断开")

    def on_stock_order(self, order):
        """委托回报推送

        :param order: XtOrder对象
            http://docs.thinktrader.net/pages/198696/#%E5%A7%94%E6%89%98xtorder
            http://docs.thinktrader.net/pages/198696/#%E5%A7%94%E6%89%98%E7%8A%B6%E6%80%81-order-status
        """
        order_status_map = {
            xtconstant.ORDER_UNREPORTED: "未报",
            xtconstant.ORDER_WAIT_REPORTING: "待报",
            xtconstant.ORDER_REPORTED: "已报",
            xtconstant.ORDER_REPORTED_CANCEL: "已报待撤",
            xtconstant.ORDER_PARTSUCC_CANCEL: "部成待撤",
            xtconstant.ORDER_PART_CANCEL: "部撤",
            xtconstant.ORDER_CANCELED: "已撤",
            xtconstant.ORDER_PART_SUCC: "部成",
            xtconstant.ORDER_SUCCEEDED: "已成",
            xtconstant.ORDER_JUNK: "废单",
            xtconstant.ORDER_UNKNOWN: "未知",
        }

        msg = (
            f"委托回报通知：\n{'*' * 31}\n"
            f"时间：{datetime.now().strftime(dt_fmt)}\n"
            f"账户：{order.account_id}\n"
            f"标的：{order.stock_code}\n"
            f"方向：{'做多' if order.order_type == 23 else '平多'}\n"
            f"数量：{int(order.order_volume)}\n"
            f"价格：{order.price}\n"
            f"状态：{order_status_map[order.order_status]}"
        )
        self.logger.info(f"on order callback: {msg}")
        if self.feishu_push_mode == "detail" and is_trade_time():
            self.push_message(msg, msg_type="text")

    def on_stock_asset(self, asset):
        """资金变动推送

        :param asset: XtAsset对象
        """
        msg = (
            f"资金变动通知: \n{'*' * 31}\n"
            f"时间：{datetime.now().strftime(dt_fmt)}\n"
            f"账户: {asset.account_id} \n"
            f"可用资金：{asset.cash} \n"
            f"总资产：{asset.total_asset}"
        )
        self.logger.info(f"on asset callback: {msg}")
        if self.feishu_push_mode == "detail" and is_trade_time():
            self.push_message(msg, msg_type="text")

    def on_stock_trade(self, trade):
        """成交变动推送

        :param trade: XtTrade对象
        """
        msg = (
            f"成交变动通知：\n{'*' * 31}\n"
            f"时间：{datetime.now().strftime(dt_fmt)}\n"
            f"账户：{trade.account_id}\n"
            f"标的：{trade.stock_code}\n"
            f"方向：{'开多' if trade.order_type == 23 else '平多'}\n"
            f"成交量：{int(trade.traded_volume)}\n"
            f"成交价：{round(trade.traded_price, 2)}"
        )
        self.logger.info(f"on trade callback: {msg}")
        if self.feishu_push_mode == "detail" and is_trade_time():
            self.push_message(msg, msg_type="text")

    def on_stock_position(self, position):
        """持仓变动推送

        :param position: XtPosition对象
        """
        msg = (
            f"持仓变动通知: \n{'*' * 31}\n"
            f"时间：{datetime.now().strftime(dt_fmt)}\n"
            f"账户：{position.account_id}\n"
            f"标的：{position.stock_code}\n"
            f"成交量：{position.volume}"
        )
        self.logger.info(f"on position callback: {msg}")
        if self.feishu_push_mode == "detail" and is_trade_time():
            self.push_message(msg, msg_type="text")

    def on_order_error(self, order_error):
        """委托失败推送

        :param order_error:XtOrderError 对象
        """
        msg = (
            f"委托失败通知: \n{'*' * 31}\n"
            f"时间：{datetime.now().strftime(dt_fmt)}\n"
            f"账户：{order_error.account_id}\n"
            f"订单编号：{order_error.order_id}\n"
            f"错误编码：{order_error.error_id}\n"
            f"失败原因：{order_error.error_msg}"
        )
        self.logger.info(f"on order_error callback: {msg}")
        if is_trade_time():
            self.push_message(msg, msg_type="text")

    def on_cancel_error(self, cancel_error):
        """撤单失败推送

        :param cancel_error: XtCancelError 对象
        """
        msg = (
            f"撤单失败通知: \n{'*' * 31}\n"
            f"时间：{datetime.now().strftime(dt_fmt)}\n"
            f"账户：{cancel_error.account_id}\n"
            f"订单编号：{cancel_error.order_id}\n"
            f"错误编码：{cancel_error.error_id}\n"
            f"失败原因：{cancel_error.error_msg}"
        )
        self.logger.info(f"on_cancel_error: {msg}")
        if is_trade_time():
            self.push_message(msg, msg_type="text")

    def on_order_stock_async_response(self, response):
        """异步下单回报推送

        :param response: XtOrderResponse 对象
        """
        msg = (
            f"异步下单回报推送: \n{'*' * 31}\n"
            f"时间：{datetime.now().strftime(dt_fmt)}\n"
            f"账户：{response.account_id}\n"
            f"订单编号：{response.order_id}\n"
            f"策略名称：{response.strategy_name}"
        )
        if is_trade_time():
            self.push_message(msg, msg_type="text")
        self.logger.info(f"on_order_stock_async_response: {msg}")

    def on_account_status(self, status):
        """账户状态变化推送

        :param status: XtAccountStatus 对象
        """
        status_map = {
            xtconstant.ACCOUNT_STATUS_OK: "正常",
            xtconstant.ACCOUNT_STATUS_WAITING_LOGIN: "连接中",
            xtconstant.ACCOUNT_STATUSING: "登陆中",
            xtconstant.ACCOUNT_STATUS_FAIL: "失败",
        }
        msg = (
            f"账户状态变化推送：\n{'*' * 31}\n"
            f"时间：{datetime.now().strftime(dt_fmt)}\n"
            f"账户ID：{status.account_id}\n"
            f"账号类型：{'证券账户' if status.account_type == 2 else '其他'}\n"
            f"账户状态：{status_map[status.status]}\n"
        )

        self.logger.info(
            f"账户ID: {status.account_id} "
            f"账号类型：{'证券账户' if status.account_type == 2 else '其他'} "
            f"账户状态：{status_map[status.status]}"
        )

        if is_trade_time():
            self.push_message(msg, msg_type="text")


def query_stock_positions(xtt: XtQuantTrader, acc: StockAccount):
    """查询股票市场的持仓单

    http://docs.thinktrader.net/pages/ee0e9b/#%E6%8C%81%E4%BB%93%E6%9F%A5%E8%AF%A2
    http://docs.thinktrader.net/pages/198696/#%E6%8C%81%E4%BB%93xtposition
    """
    res = xtt.query_stock_positions(acc)
    res = {x.stock_code: x for x in res} if len(res) > 0 else {}
    return res


def query_today_trades(xtt: XtQuantTrader, acc: StockAccount):
    """查询当日成交

    http://docs.thinktrader.net/pages/198696/#%E6%88%90%E4%BA%A4xttrade
    """
    trades = xtt.query_stock_trades(acc)
    res = [
        {
            "品种": x.stock_code,
            "均价": x.traded_price,
            "方向": "买入" if x.order_type == 23 else "卖出",
            "数量": x.traded_volume,
            "金额": x.traded_amount,
            "时间": time.strftime("%H:%M:%S", time.localtime(x.traded_time)),
        }
        for x in trades
    ]
    return res


def cancel_timeout_orders(xtt: XtQuantTrader, acc: StockAccount, minutes=30, **kwargs):
    """撤销超时的委托单

    http://docs.thinktrader.net/pages/ee0e9b/#%E8%82%A1%E7%A5%A8%E5%90%8C%E6%AD%A5%E6%92%A4%E5%8D%95
    http://docs.thinktrader.net/pages/ee0e9b/#%E5%A7%94%E6%89%98%E6%9F%A5%E8%AF%A2
    http://docs.thinktrader.net/pages/198696/#%E5%A7%94%E6%89%98xtorder

    :param minutes: 超时时间，单位分钟
    :return:
    """
    logger = kwargs.get("logger", loguru.logger)
    orders = xtt.query_stock_orders(acc, cancelable_only=True)
    for o in orders:
        if datetime.fromtimestamp(o.order_time) < datetime.now() - timedelta(minutes=minutes):
            xtt.cancel_order_stock(acc, o.order_id)
            logger.info(f"撤销超时委托单：{o.order_id}; {o.stock_code}; {o.order_volume}; {o.order_type}")


def query_stock_orders(xtt, acc, **kwargs):
    """查询股票委托单

    :param xtt: XtQuantTrader, QMT 交易接口
    :param acc: StockAccount, 账户
    :param kwargs:
        cancelable_only: bool, 是否只查询可撤单的委托单
        start_time: str, 开始时间，格式：09:00:00
    """
    cancelable_only = kwargs.get("cancelable_only", False)
    orders = xtt.query_stock_orders(acc, cancelable_only)

    start_time = kwargs.get("start_time", "09:00:00")
    rows = []
    for order in orders:
        row = {
            "account_id": order.account_id,
            "stock_code": order.stock_code,
            "order_id": order.order_id,
            "order_sysid": order.order_sysid,
            "order_time": order.order_time,
            "order_type": order.order_type,
            "order_volume": order.order_volume,
            "price_type": order.price_type,
            "price": order.price,
            "traded_volume": order.traded_volume,
            "traded_price": order.traded_price,
            "order_status": order.order_status,
            "status_msg": order.status_msg,
            "strategy_name": order.strategy_name,
            "order_remark": order.order_remark,
            "direction": order.direction,
            "offset_flag": order.offset_flag,
        }
        rows.append(row)

    dfr = pd.DataFrame(rows)
    dfr["order_time"] = pd.to_datetime(dfr["order_time"], unit="s") + pd.Timedelta(hours=8)
    dfr["order_date"] = dfr["order_time"].dt.strftime("%Y-%m-%d")
    dfr["order_time"] = dfr["order_time"].dt.strftime("%H:%M:%S")
    if start_time:
        dfr = dfr[dfr["order_time"] >= start_time].copy().reset_index(drop=True)
    return dfr


def is_order_exist(xtt: XtQuantTrader, acc: StockAccount, symbol, order_type, volume=None):
    """判断是否存在相同方向的委托单"""
    orders = xtt.query_stock_orders(acc, cancelable_only=False)
    for o in orders:
        if o.stock_code == symbol and o.order_type == order_type:
            if not volume or o.order_volume == volume:
                return True
    return False


def is_allow_open(xtt: XtQuantTrader, acc: StockAccount, symbol, **kwargs):
    """判断是否允许开仓

    http://dict.thinktrader.net/nativeApi/xttrader.html?id=H018C2#%E8%B5%84%E4%BA%A7xtasset

    :param xtt: XtQuantTrader, QMT 交易接口
    :param acc: StockAccount, 账户
    :param symbol: 股票代码
    :param kwargs:

        forbidden_symbols: list, 禁止交易的标的
        multiple_orders: bool, 是否允许多次下单

    :return: True 允许开仓，False 不允许开仓
    """
    logger = kwargs.get("logger", loguru.logger)

    # 如果 symbol 在禁止交易的列表中，不允许开仓
    if symbol in kwargs.get("forbidden_symbols", []):
        return False

    # 如果 未成交的开仓委托单 存在，不允许开仓
    multiple_orders = kwargs.get("multiple_orders", False)
    if not multiple_orders:
        if is_order_exist(xtt, acc, symbol, order_type=23):
            logger.warning(f"存在未成交的开仓委托单，symbol={symbol}")
            return False

    # 如果资金不足，不允许开仓
    assets = xtt.query_stock_asset(acc)
    symbol_price = xtdata.get_full_tick([symbol])[symbol]["lastPrice"]
    if assets.cash < symbol_price * 120:
        logger.warning(f"资金不足，无法开仓，symbol={symbol}")
        return False

    return True


def is_allow_exit(xtt: XtQuantTrader, acc: StockAccount, symbol, **kwargs):
    """判断是否允许平仓

    :param xtt: XtQuantTrader, QMT 交易接口
    :param acc: StockAccount, 账户
    :param symbol: 股票代码
    :param kwargs:

        forbidden_symbols: list, 禁止交易的标的
        multiple_orders: bool, 是否允许多次下单

    :return: True 允许开仓，False 不允许开仓
    """
    logger = kwargs.get("logger", loguru.logger)

    # symbol 在禁止交易的列表中，不允许平仓
    if symbol in kwargs.get("forbidden_symbols", []):
        return False

    # 没有持仓 或 可用数量为 0，不允许平仓
    pos = query_stock_positions(xtt, acc).get(symbol, None)
    if not pos or pos.can_use_volume <= 0:
        logger.warning(f"没有持仓或可用数量为0，无法平仓，symbol={symbol}")
        return False

    multiple_orders = kwargs.get("multiple_orders", False)
    if not multiple_orders:
        # 未成交的平仓委托单 存在，不允许继续平仓
        if is_order_exist(xtt, acc, symbol, order_type=24):
            logger.warning(f"存在未成交的平仓委托单，symbol={symbol}")
            return False

    return True


def send_stock_order(xtt: XtQuantTrader, acc: StockAccount, **kwargs):
    """股票市场交易下单

    股票同步报单 http://docs.thinktrader.net/pages/ee0e9b/#%E8%82%A1%E7%A5%A8%E5%90%8C%E6%AD%A5%E6%8A%A5%E5%8D%95
    委托类型(order_type) http://docs.thinktrader.net/pages/198696/#%E5%A7%94%E6%89%98%E7%B1%BB%E5%9E%8B-order-type
    报价类型(price_type) http://docs.thinktrader.net/pages/198696/#%E6%8A%A5%E4%BB%B7%E7%B1%BB%E5%9E%8B-price-type

    stock_code: 证券代码, 例如"600000.SH"
    order_type: 委托类型, 23:买, 24:卖
    order_volume: 委托数量, 股票以'股'为单位, 债券以'张'为单位，ETF以'份'为单位；数量必须是100的整数倍
    price_type: 报价类型, 详见帮助手册
        xtconstant.LATEST_PRICE	    5	最新价
        xtconstant.FIX_PRICE	    11	限价
    price: 报价价格, 如果price_type为限价, 那price为指定的价格, 否则填0
    strategy_name: 策略名称
    order_remark: 委托备注

    :return: 返回下单请求序号, 成功委托后的下单请求序号为大于0的正整数, 如果为-1表示委托失败
    """
    logger = kwargs.get("logger", loguru.logger)

    stock_code = kwargs.get("stock_code")
    order_type = kwargs.get("order_type")
    order_volume = kwargs.get("order_volume")  # 委托数量, 股票以'股'为单位, 债券以'张'为单位
    price_type = kwargs.get("price_type", 5)
    price = kwargs.get("price", 0)
    strategy_name = kwargs.get("strategy_name", "程序下单")
    order_remark = kwargs.get("order_remark", "程序下单")

    if not xtt.connected:
        xtt.start()
        xtt.connect()

    order_volume = max(order_volume // 100 * 100, 0)  # 股票市场只允许做多 100 的整数倍
    assert xtt.connected, "交易服务器连接断开"
    _id = xtt.order_stock(
        acc, stock_code, order_type, int(order_volume), price_type, price, strategy_name, order_remark
    )
    return _id


def order_stock_target(xtt: XtQuantTrader, acc: StockAccount, symbol, target, **kwargs):
    """下单调整至目标仓位

    :param xtt: XtQuantTrader, QMT 交易接口
    :param acc: StockAccount, 账户
    :param symbol: str, 股票代码
    :param target: int, 目标仓位, 单位：股；正数表示持有多头仓位，负数表示持有空头仓位
    :param kwargs: dict, 其他参数

        - price_type: int, 报价类型, 详见帮助手册
            xtconstant.LATEST_PRICE	    5	最新价
            xtconstant.FIX_PRICE	    11	限价
        - price: float, 报价价格, 如果price_type为限价, 那price为指定的价格, 否则填0
        - logger: loguru.logger, 日志记录器
        - forbidden_symbols: list, 禁止交易的标的
        - multiple_orders: bool, 是否允许多次下单

    :return:
    """
    logger = kwargs.get("logger", loguru.logger)

    # 查询持仓
    pos = query_stock_positions(xtt, acc).get(symbol, None)
    current = pos.volume if pos else 0

    logger.info(f"当前持仓：{current}，目标仓位：{target}")
    if current == target:
        return

    price_type = kwargs.get("price_type", 5)
    price = kwargs.get("price", 0)

    # 如果目标小于当前，平仓
    if target < current and is_allow_exit(xtt, acc, symbol, **kwargs):
        delta = min(current - target, pos.can_use_volume if pos else current)
        logger.info(f"{symbol}平仓，目标仓位：{target}，当前仓位：{current}，平仓数量：{delta}")
        if delta != 0:
            send_stock_order(
                xtt, acc, stock_code=symbol, order_type=24, order_volume=delta, price_type=price_type, price=price
            )
            return

    # 如果目标大于当前，开仓
    if target > current and is_allow_open(xtt, acc, symbol, **kwargs):
        delta = target - current
        logger.info(f"{symbol}开仓，目标仓位：{target}，当前仓位：{current}，开仓数量：{delta}")
        if delta != 0:
            send_stock_order(
                xtt, acc, stock_code=symbol, order_type=23, order_volume=delta, price_type=price_type, price=price
            )
            return


class QmtTradeManager:
    """QMT交易管理器（这是一个案例性质的存在，真正实盘的时候请参考这个，根据自己的逻辑重新实现）

    功能特性：

    1. 支持全市场品种定时交易
    2. 所有交易对象持久化，程序重启后自动恢复
    3. 仅支持股票交易
    4. 仅在仓位发生变化时进行交易

    """

    def __init__(self, mini_qmt_dir, account_id, **kwargs):
        """

        :param mini_qmt_dir: mini QMT 路径；如 D:\\国金QMT交易端模拟\\userdata_mini
        :param account_id: 账户ID
        :param kwargs:

        """
        self.logger = kwargs.get("logger", loguru.logger)

        self.cache_path = kwargs["cache_path"]  # 交易缓存路径
        os.makedirs(self.cache_path, exist_ok=True)
        self.symbols = kwargs.get("symbols", [])  # 交易标的列表
        self.strategy = kwargs.get("strategy", [])  # 交易策略
        assert issubclass(self.strategy, czsc.CzscStrategyBase), "交易策略必须是CzscStrategyBase的子类"

        self.symbol_max_pos = kwargs.get("symbol_max_pos", 0.5)  # 每个标的最大持仓比例
        self.trade_sdt = kwargs.get("trade_sdt", "20220601")  # 交易跟踪开始日期
        self.mini_qmt_dir = mini_qmt_dir
        self.account_id = account_id
        self.base_freq = self.strategy(symbol="symbol").sorted_freqs[0]
        self.delta_days = int(kwargs.get("delta_days", 1))  # 定时执行获取的K线天数
        self.forbidden_symbols = kwargs.get("forbidden_symbols", [])  # 禁止交易的品种列表

        self.session = random.randint(10000, 20000)
        self.callback = TraderCallback(**kwargs.get("callback_params", {}))
        self.xtt = XtQuantTrader(mini_qmt_dir, session=self.session, callback=self.callback)
        self.acc = StockAccount(account_id, "STOCK")
        self.xtt.start()
        self.xtt.connect()
        assert self.xtt.connected, "交易服务器连接失败"
        _res = self.xtt.subscribe(self.acc)
        assert _res == 0, "账号订阅失败"
        self.traders = self.__create_traders(**kwargs)

    def __create_traders(self, **kwargs):
        """创建交易策略"""
        traders = {}
        for symbol in tqdm(self.symbols, desc="创建交易对象", unit="个"):
            if symbol in self.forbidden_symbols:
                continue

            file_trader = os.path.join(self.cache_path, f"{symbol}.ct")
            try:
                if os.path.exists(file_trader):
                    # 从缓存文件中恢复交易对象，并更新K线数据
                    trader: CzscTrader = czsc.dill_load(file_trader)
                    kline_sdt = pd.to_datetime(trader.end_dt) - timedelta(days=self.delta_days)
                    bars = get_raw_bars(
                        symbol, self.base_freq, kline_sdt, datetime.now(), fq="前复权", download_hist=True
                    )
                    news = [x for x in bars if x.dt > trader.end_dt]
                    if news:
                        self.logger.info(f"{symbol} 需要更新的K线数量：{len(news)} | 最新的K线时间是 {news[-1].dt}")
                        for bar in news:
                            trader.on_bar(bar)

                else:
                    # 从头创建交易对象
                    bars = get_raw_bars(symbol, self.base_freq, "20180101", datetime.now(), fq="前复权")
                    trader: CzscTrader = self.strategy(symbol=symbol).init_trader(bars, sdt=self.trade_sdt)
                    czsc.dill_dump(trader, file_trader)

                mean_pos = trader.get_ensemble_pos("mean")
                if mean_pos == 0:
                    continue

                traders[symbol] = trader
                pos_info = {x.name: x.pos for x in trader.positions if x.pos != 0}
                self.logger.info(f"最新时间：{trader.end_dt}；{symbol} trader pos：{pos_info} | mean_pos: {mean_pos}")
            except Exception as e:
                self.logger.exception(f"创建交易对象失败，symbol={symbol}, e={e}")

        return traders

    def get_assets(self):
        """获取账户资产"""
        return self.xtt.query_stock_asset(self.acc)

    def query_stock_orders(self, cancelable_only=False):
        """查询股票市场的委托单

        http://docs.thinktrader.net/pages/ee0e9b/#%E5%A7%94%E6%89%98%E6%9F%A5%E8%AF%A2

        :param cancelable_only:
        :return:
        """
        return self.xtt.query_stock_orders(self.acc, cancelable_only)

    def query_today_trades(self):
        """查询当日成交"""
        # http://docs.thinktrader.net/pages/198696/#%E6%88%90%E4%BA%A4xttrade
        trades = self.xtt.query_stock_trades(self.acc)
        res = [
            {
                "品种": x.stock_code,
                "均价": x.traded_price,
                "方向": "买入" if x.order_type == 23 else "卖出",
                "数量": x.traded_volume,
                "金额": x.traded_amount,
                "时间": time.strftime("%H:%M:%S", time.localtime(x.traded_time)),
            }
            for x in trades
        ]
        return res

    def cancel_timeout_orders(self, minutes=30):
        """撤销超时的委托单

        :param minutes: 超时时间，单位分钟
        :return:
        """
        orders = self.query_stock_orders(cancelable_only=True)
        for o in orders:
            if datetime.fromtimestamp(o.order_time) < datetime.now() - timedelta(minutes=minutes):
                self.xtt.cancel_order_stock(self.acc, o.order_id)

    def is_order_exist(self, symbol, order_type, volume=None):
        """判断是否存在相同的委托单"""
        orders = self.query_stock_orders(cancelable_only=True)
        for o in orders:
            if o.stock_code == symbol and o.order_type == order_type:
                if not volume or o.order_volume == volume:
                    return True
        return False

    def is_allow_open(self, symbol, price):
        """判断是否允许开仓

        :param symbol: 股票代码
        :param price: 股票现价
        :return: True 允许开仓，False 不允许开仓
        """
        # 如果 symbol 在禁止交易的列表中，不允许开仓
        if symbol in self.forbidden_symbols:
            return False

        # 如果 未成交的开仓委托单 存在，不允许开仓
        if self.is_order_exist(symbol, order_type=23):
            self.logger.warning(f"存在未成交的开仓委托单，symbol={symbol}")
            return False

        # 如果 symbol_max_pos 为 0，不允许开仓
        if self.symbol_max_pos <= 0:
            return False

        # 如果已经有持仓，不允许开仓
        if self.query_stock_positions().get(symbol, None):
            return False

        # 如果资金不足，不允许开仓
        assets = self.get_assets()
        if assets.cash < price * 120:
            self.logger.warning(f"资金不足，无法开仓，symbol={symbol}")
            return False

        return True

    def is_allow_exit(self, symbol):
        """判断是否允许平仓

        :param symbol: 股票代码
        :return: True 允许开仓，False 不允许开仓
        """
        # symbol 在禁止交易的列表中，不允许平仓
        if symbol in self.forbidden_symbols:
            return False

        # 没有持仓，不允许平仓
        pos = self.query_stock_positions().get(symbol)
        if not pos:
            return False

        # 未成交的平仓委托单 存在，不允许平仓
        if self.is_order_exist(symbol, order_type=24):
            self.logger.warning(f"存在未成交的平仓委托单，symbol={symbol}")
            return False

        # 持仓可用数量为 0，不允许平仓
        if pos.can_use_volume <= 0:
            return False

        return True

    def query_stock_positions(self):
        """查询股票市场的持仓单

        http://docs.thinktrader.net/pages/ee0e9b/#%E6%8C%81%E4%BB%93%E6%9F%A5%E8%AF%A2
        """
        res = self.xtt.query_stock_positions(self.acc)
        if len(res) > 0:
            res = {x.stock_code: x for x in res}
        else:
            res = {}
        return res

    def send_stock_order(self, **kwargs):
        """股票市场交易下单

        http://docs.thinktrader.net/pages/ee0e9b/#%E8%82%A1%E7%A5%A8%E5%90%8C%E6%AD%A5%E6%8A%A5%E5%8D%95
        http://docs.thinktrader.net/pages/198696/#%E6%8A%A5%E4%BB%B7%E7%B1%BB%E5%9E%8B-price-type

        stock_code: 证券代码, 例如"600000.SH"
        order_type: 委托类型, 23:买, 24:卖
        order_volume: 委托数量, 股票以'股'为单位, 债券以'张'为单位
        price_type: 报价类型, 详见帮助手册
            xtconstant.LATEST_PRICE	5	最新价
            xtconstant.FIX_PRICE	11	限价
        price: 报价价格, 如果price_type为限价, 那price为指定的价格, 否则填0
        strategy_name: 策略名称
        order_remark: 委托备注

        :return: 返回下单请求序号, 成功委托后的下单请求序号为大于0的正整数, 如果为-1表示委托失败
        """
        stock_code = kwargs.get("stock_code")
        order_type = kwargs.get("order_type")
        order_volume = kwargs.get("order_volume")  # 委托数量, 股票以'股'为单位, 债券以'张'为单位
        price_type = kwargs.get("price_type", xtconstant.LATEST_PRICE)
        price = kwargs.get("price", 0)
        strategy_name = kwargs.get("strategy_name", "程序下单")
        order_remark = kwargs.get("order_remark", "程序下单")

        if not self.xtt.connected:
            self.xtt.connect()
            self.xtt.start()

        if order_volume % 100 != 0:
            order_volume = order_volume // 100 * 100

        assert self.xtt.connected, "交易服务器连接断开"
        _id = self.xtt.order_stock(
            self.acc, stock_code, order_type, int(order_volume), price_type, price, strategy_name, order_remark
        )
        return _id

    def update_traders(self):
        """更新交易策略"""
        holds = self.query_stock_positions()
        kline_sdt = datetime.now() - timedelta(days=self.delta_days)

        for symbol in self.traders.keys():
            try:
                trader = self.traders[symbol]
                bars = get_raw_bars(symbol, self.base_freq, kline_sdt, datetime.now(), fq="前复权", download_hist=True)

                news = [x for x in bars if x.dt > trader.end_dt]
                if news:
                    self.logger.info(f"{symbol} 需要更新的K线数量：{len(news)} | 最新的K线时间是 {news[-1].dt}")
                    for bar in news:
                        trader.on_bar(bar)

                        # 根据策略的交易信号，下单【股票只有多头】，只有当信号变化时才下单
                        if (
                            trader.get_ensemble_pos(method="vote") == 1
                            and trader.pos_changed
                            and self.is_allow_open(symbol, price=news[-1].close)
                        ):
                            assets = self.get_assets()
                            order_volume = min(self.symbol_max_pos * assets.total_asset, assets.cash) // news[-1].close
                            self.send_stock_order(stock_code=symbol, order_type=23, order_volume=order_volume)

                        # 平多头
                        if trader.get_ensemble_pos(method="vote") == 0 and self.is_allow_exit(symbol):
                            order_volume = holds[symbol].can_use_volume
                            self.send_stock_order(stock_code=symbol, order_type=24, order_volume=order_volume)

                else:
                    self.logger.info(f"{symbol} 没有需要更新的K线，最新的K线时间是 {trader.end_dt}")

                if trader.get_ensemble_pos("mean") > 0:
                    pos_info = {x.name: x.pos for x in trader.positions if x.pos != 0}
                    self.logger.info(
                        f"{trader.end_dt} {symbol} trader pos：{pos_info} | ensemble_pos: {trader.get_ensemble_pos('mean')}"
                    )

                # 更新交易对象
                self.traders[symbol] = trader

            except Exception as e:
                self.callback.push_message(f"{symbol} 更新交易策略失败，原因是 {e}")
                self.logger.error(f"{symbol} 更新交易策略失败，原因是 {e}")

    def update_offline_traders(self):
        """更新全部品种策略"""
        traders = {}
        kline_sdt = datetime.now() - timedelta(days=self.delta_days)

        for symbol in tqdm(self.symbols, desc="更新全部股票", unit="个"):
            if symbol in self.forbidden_symbols:
                continue

            file_trader = os.path.join(self.cache_path, f"{symbol}.ct")
            if not os.path.exists(file_trader):
                self.logger.error(f"{symbol} 交易对象不存在，无法更新")
                continue

            try:
                bars = get_raw_bars(symbol, self.base_freq, kline_sdt, datetime.now(), fq="前复权", download_hist=True)
                trader: CzscTrader = czsc.dill_load(file_trader)
                news = [x for x in bars if x.dt > trader.end_dt]
                if news:
                    self.logger.info(f"{symbol} 需要更新的K线数量：{len(news)} | 最新的K线时间是 {news[-1].dt}")
                    for bar in news:
                        trader.on_bar(bar)

                        # 根据策略的交易信号，下单【股票只有多头】，只有当信号变化时才下单
                        if (
                            trader.get_ensemble_pos(method="vote") == 1
                            and trader.pos_changed
                            and self.is_allow_open(symbol, price=news[-1].close)
                        ):
                            assets = self.get_assets()
                            order_volume = min(self.symbol_max_pos * assets.total_asset, assets.cash) // news[-1].close
                            self.send_stock_order(stock_code=symbol, order_type=23, order_volume=order_volume)

                    czsc.dill_dump(trader, file_trader)

                mean_pos = trader.get_ensemble_pos("mean")
                if mean_pos == 0:
                    continue

                traders[symbol] = trader
                pos_info = {x.name: x.pos for x in trader.positions if x.pos != 0}
                self.logger.info(f"最新时间：{trader.end_dt}；{symbol} trader pos：{pos_info} | mean_pos: {mean_pos}")
            except Exception as e:
                self.logger.exception(f"创建交易对象失败，symbol={symbol}, e={e}")

        self.traders = traders

    def report(self):
        """报告状态"""
        from czsc.utils import WordWriter

        writer = WordWriter()
        writer.add_title("QMT 交易报告")
        assets = self.get_assets()

        writer.add_heading("一、账户状态", level=1)
        writer.add_paragraph(
            f"交易品种数量：{len(self.traders)}\n"
            f"传入品种数量：{len(self.symbols)}\n"
            f"交易账户：{self.account_id}\n"
            f"账户资产：{assets.total_asset}\n"
            f"可用资金：{assets.cash}\n"
            f"持仓市值：{assets.market_value}\n"
            f"持仓情况：",
            first_line_indent=0,
        )

        sp = self.query_stock_positions()
        if sp:
            _res_sp = []
            for k, v in sp.items():
                is_auto = "程序" if k in self.traders.keys() else "人工"
                _res_sp.append(
                    {
                        "品种": k,
                        "持仓股数": v.volume,
                        "可用股数": v.can_use_volume,
                        "成本": v.open_price,
                        "市值": int(v.market_value),
                        "操盘手": is_auto,
                    }
                )
            writer.add_df_table(pd.DataFrame(_res_sp))
        else:
            writer.add_paragraph("当前没有持仓", first_line_indent=0)

        # 当日成交
        trades = self.query_today_trades()
        writer.add_paragraph(f"当日成交：", first_line_indent=0)
        if trades:
            writer.add_df_table(pd.DataFrame(trades))
        else:
            writer.add_paragraph("当日没有成交", first_line_indent=0)

        writer.add_heading("二、策略状态", level=1)

        _res = []
        for symbol, trader in self.traders.items():
            if trader.get_ensemble_pos("mean") > 0:
                _pos_str = "\n\n".join([f"{x.name}：{x.pos}" for x in trader.positions if x.pos != 0])
                _ops = [x.operates[-1] for x in trader.positions if x.pos != 0]
                _ops_str = "\n\n".join([f"时间：{x['dt']}_价格：{x['price']}_描述：{x['op_desc']}" for x in _ops])
                _res.append(
                    {
                        "symbol": symbol,
                        "pos": round(trader.get_ensemble_pos("mean"), 3),
                        "positions": _pos_str,
                        "operates": _ops_str,
                    }
                )
        if _res:
            writer.add_df_table(pd.DataFrame(_res).sort_values(by="pos", ascending=False))
        else:
            writer.add_paragraph("当前所有品种都是空仓")

        file_docx = f"QMT{self.account_id}_交易报告_{datetime.now().strftime('%Y%m%d_%H%M')}.docx"
        writer.save(file_docx)
        self.callback.push_message(file_docx, msg_type="file")
        os.remove(file_docx)

    def run(self, mode="30m", order_timeout=120):
        """运行策略"""
        self.report()

        if mode.lower() == "15m":
            _times = [
                "09:45",
                "10:00",
                "10:15",
                "10:30",
                "10:45",
                "11:00",
                "11:15",
                "11:30",
                "13:15",
                "13:30",
                "13:45",
                "14:00",
                "14:15",
                "14:30",
                "14:45",
                "15:00",
            ]
        elif mode.lower() == "30m":
            _times = ["09:45", "10:00", "10:30", "11:00", "11:30", "13:30", "14:00", "14:30", "15:00"]
        elif mode.lower() == "60m":
            _times = ["10:30", "11:30", "13:45", "14:30"]
        else:
            raise ValueError("mode 只能是 15m, 30m, 60m")

        while 1:
            now_dt = datetime.now().strftime("%H:%M")
            self.cancel_timeout_orders(minutes=order_timeout)

            if is_trade_day() and now_dt in _times:
                self.update_traders()
                self.report()
                time.sleep(60)
            else:
                time.sleep(3)

            # 如果断开，重新连接交易服务器
            if not self.xtt.connected:
                self.xtt.connect()
                self.xtt.start()

            if now_dt in ["11:35", "14:05", "15:05"]:
                self.callback.push_message(f"{self.account_id} 开始更新离线交易员对象", msg_type="text")
                self.update_offline_traders()
                self.report()


# ======================================================================================================================
# 以下是测试代码
# ======================================================================================================================


def test_get_kline():
    # 获取所有板块
    slt = xtdata.get_sector_list()
    stocks = xtdata.get_stock_list_in_sector("沪深A股")

    df = get_kline(
        symbol="000001.SZ", period="1m", count=1000, dividend_type="front", start_time="20200427", end_time="20221231"
    )
    assert not df.empty
    df = get_kline(
        symbol="000001.SZ", period="5m", count=1000, dividend_type="front", start_time="20200427", end_time="20221231"
    )
    assert not df.empty
    df = get_kline(
        symbol="000001.SZ", period="1d", count=1000, dividend_type="front", start_time="20200427", end_time="20221231"
    )
    assert not df.empty


def test_get_symbols():
    symbols = get_symbols("index")
    assert len(symbols) > 0
    symbols = get_symbols("stock")
    assert len(symbols) > 0
    symbols = get_symbols("check")
    assert len(symbols) > 0
    symbols = get_symbols("train")
    assert len(symbols) > 0
    symbols = get_symbols("valid")
    assert len(symbols) > 0
    symbols = get_symbols("etfs")
    assert len(symbols) > 0



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2024/1/1 00:00
describe: czsc.utils.cache 单元测试
"""
import os
import time
import shutil
import tempfile
import pandas as pd
from pathlib import Path
from czsc.utils.cache import (
    disk_cache, home_path, empty_cache_path, get_dir_size, 
    DiskCache, clear_expired_cache, clear_cache
)

# 创建临时路径用于测试
temp_path = os.path.join(home_path, "temp")


def test_get_dir_size():
    """测试获取目录大小功能"""
    # 创建临时目录和文件
    test_dir = tempfile.mkdtemp()
    
    # 创建一个测试文件
    test_file = os.path.join(test_dir, "test.txt")
    with open(test_file, "w") as f:
        f.write("hello world")
    
    # 获取目录大小
    size = get_dir_size(test_dir)
    assert size > 0
    assert isinstance(size, int)
    
    # 清理
    shutil.rmtree(test_dir)


def test_empty_cache_path():
    """测试清空缓存路径功能"""
    # 使用临时目录来测试，避免影响其他测试
    test_dir = tempfile.mkdtemp()
    test_file = os.path.join(test_dir, "test_file.txt")
    
    # 创建测试文件
    with open(test_file, "w") as f:
        f.write("test content")
    
    # 确认文件存在
    assert os.path.exists(test_file)
    
    # 模拟清空缓存路径的逻辑（不直接调用empty_cache_path以避免影响其他测试）
    shutil.rmtree(test_dir)
    os.makedirs(test_dir, exist_ok=False)
    
    # 确认文件已被删除，但目录重新创建
    assert os.path.exists(test_dir)
    assert not os.path.exists(test_file)
    
    # 清理
    shutil.rmtree(test_dir)


def test_disk_cache_class():
    """测试 DiskCache 类的基本功能"""
    # 创建临时缓存实例
    cache = DiskCache()
    
    # 测试字符串方法
    assert "DiskCache" in str(cache)
    
    # 测试设置和获取 pkl 文件
    test_key = "test_pkl"
    test_value = {"test": "data", "number": 123}
    
    cache.set(test_key, test_value)
    assert cache.is_found(test_key)
    retrieved_value = cache.get(test_key)
    assert retrieved_value == test_value
    
    # 测试删除缓存
    cache.remove(test_key)
    assert not cache.is_found(test_key)


def test_disk_cache_json():
    """测试 DiskCache 的 JSON 格式支持"""
    cache = DiskCache()
    
    test_key = "test_json"
    test_value = {"name": "test", "value": 42}
    
    cache.set(test_key, test_value, suffix="json")
    assert cache.is_found(test_key, suffix="json")
    retrieved_value = cache.get(test_key, suffix="json")
    assert retrieved_value == test_value
    
    cache.remove(test_key, suffix="json")


def test_disk_cache_txt():
    """测试 DiskCache 的文本格式支持"""
    cache = DiskCache()
    
    test_key = "test_txt"
    test_value = "这是测试文本内容"
    
    cache.set(test_key, test_value, suffix="txt")
    assert cache.is_found(test_key, suffix="txt")
    retrieved_value = cache.get(test_key, suffix="txt")
    assert retrieved_value == test_value
    
    cache.remove(test_key, suffix="txt")


def test_disk_cache_dataframe():
    """测试 DiskCache 的 DataFrame 格式支持"""
    cache = DiskCache()
    
    test_key = "test_df"
    test_df = pd.DataFrame({
        'A': [1, 2, 3],
        'B': ['a', 'b', 'c'],
        'C': [1.1, 2.2, 3.3]
    })
    
    # 测试 CSV 格式
    cache.set(test_key, test_df, suffix="csv")
    assert cache.is_found(test_key, suffix="csv")
    retrieved_df = cache.get(test_key, suffix="csv")
    pd.testing.assert_frame_equal(test_df, retrieved_df)
    cache.remove(test_key, suffix="csv")
    
    # 测试 Excel 格式
    cache.set(test_key, test_df, suffix="xlsx")
    assert cache.is_found(test_key, suffix="xlsx")
    retrieved_df = cache.get(test_key, suffix="xlsx")
    pd.testing.assert_frame_equal(test_df, retrieved_df)
    cache.remove(test_key, suffix="xlsx")
    
    # 测试 Feather 格式
    cache.set(test_key, test_df, suffix="feather")
    assert cache.is_found(test_key, suffix="feather")
    retrieved_df = cache.get(test_key, suffix="feather")
    pd.testing.assert_frame_equal(test_df, retrieved_df)
    cache.remove(test_key, suffix="feather")
    
    # 测试 Parquet 格式
    cache.set(test_key, test_df, suffix="parquet")
    assert cache.is_found(test_key, suffix="parquet")
    retrieved_df = cache.get(test_key, suffix="parquet")
    pd.testing.assert_frame_equal(test_df, retrieved_df)
    cache.remove(test_key, suffix="parquet")


def test_disk_cache_ttl():
    """测试 DiskCache 的 TTL 功能"""
    cache = DiskCache()
    
    test_key = "test_ttl"
    test_value = "ttl测试数据"
    
    # 设置短TTL
    cache.set(test_key, test_value)
    assert cache.is_found(test_key, ttl=1)  # 1秒TTL
    
    # 等待超时
    time.sleep(2)
    assert not cache.is_found(test_key, ttl=1)


# 创建用于装饰器测试的简单函数
@disk_cache(path=temp_path, suffix="pkl", ttl=100)
def run_func_x(x):
    return x * 2


@disk_cache(path=temp_path, suffix="txt", ttl=100)
def run_func_text(x):
    return f"hello {x}"


@disk_cache(path=temp_path, suffix="json", ttl=100)
def run_func_json(x):
    return {"a": 1, "b": 2, "x": x}


@disk_cache(path=temp_path, suffix="xlsx", ttl=100)
def run_func_y(x):
    df = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6], 'x': [x, x, x]})
    return df


@disk_cache(path=temp_path, suffix="feather", ttl=100)
def run_feather(x):
    df = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6], 'x': [x, x, x]})
    return df


@disk_cache(path=temp_path, suffix="parquet", ttl=100)
def run_parquet(x):
    df = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6], 'x': [x, x, x]})
    return df


def test_disk_cache_decorator():
    """测试 disk_cache 装饰器功能"""
    # 确保缓存根目录存在
    os.makedirs(temp_path, exist_ok=True)
    
    # 调用函数
    result = run_func_x(5)
    
    # 检查输出是否正确
    assert result == 10

    # 再次调用相同参数的函数
    result = run_func_x(5)
    
    # 检查输出仍然正确
    assert result == 10

    # 调用不同参数的函数
    result = run_func_text(6)
    result = run_func_text(6)
    assert result == "hello 6"

    # 调用 JSON 函数
    result = run_func_json(7)
    result = run_func_json(7)
    assert result == {"a": 1, "b": 2, "x": 7}

    # 调用 Feather 函数
    result = run_feather(8)
    result = run_feather(8)
    assert isinstance(result, pd.DataFrame)

    # 调用 Parquet 函数
    result = run_parquet(9)
    result = run_parquet(9)
    assert isinstance(result, pd.DataFrame)

    # 检查缓存文件是否存在
    files = os.listdir(os.path.join(temp_path, "run_func_x"))
    assert len(files) == 1

    # 调用不同参数的函数
    result = run_func_y(5)
    files = os.listdir(os.path.join(temp_path, "run_func_y"))
    assert len(files) == 1
    file_xlsx = [x for x in files if x.endswith("xlsx")][0]
    df = pd.read_excel(os.path.join(temp_path, f"run_func_y/{file_xlsx}"))
    assert isinstance(df, pd.DataFrame)


def test_clear_expired_cache():
    """测试清理过期缓存功能"""
    # 创建临时目录
    test_dir = tempfile.mkdtemp()
    
    # 创建一个pkl文件
    old_file = os.path.join(test_dir, "old_file.pkl")
    
    with open(old_file, "w") as f:
        f.write("old content")
    
    # 修改文件的修改时间为很久以前
    old_time = time.time() - 3600 * 24 * 31  # 31天前
    os.utime(old_file, (old_time, old_time))
    
    # 检查文件修改时间
    current_time = time.time()
    old_file_stat = os.stat(old_file)
    age = current_time - old_file_stat.st_mtime
    max_age = 3600 * 24 * 30  # 30天
    
    print(f"当前时间: {current_time}")
    print(f"文件修改时间: {old_file_stat.st_mtime}")
    print(f"文件年龄: {age} 秒 ({age / (3600 * 24):.1f} 天)")
    print(f"最大年龄: {max_age} 秒 ({max_age / (3600 * 24):.1f} 天)")
    print(f"是否过期: {age > max_age}")
    
    # 确保文件存在
    assert os.path.exists(old_file)
    
    # 运行清理功能
    clear_expired_cache(test_dir, max_age=max_age)
    
    # 现在应该删除过期文件
    assert not os.path.exists(old_file)
    
    # 清理
    shutil.rmtree(test_dir)


def test_clear_cache():
    """测试清理缓存功能"""
    # 创建临时目录
    test_dir = tempfile.mkdtemp()
    
    # 创建子目录和文件
    sub1 = os.path.join(test_dir, "sub1")
    sub2 = os.path.join(test_dir, "sub2")
    os.makedirs(sub1)
    os.makedirs(sub2)
    
    with open(os.path.join(sub1, "file1.txt"), "w") as f:
        f.write("content1")
    with open(os.path.join(sub2, "file2.txt"), "w") as f:
        f.write("content2")
    
    # 清理特定子目录
    clear_cache(test_dir, subs=["sub1"], recreate=True)
    
    # 检查结果
    assert os.path.exists(sub1)  # 重新创建了
    assert os.path.exists(sub2)  # 保留
    assert not os.path.exists(os.path.join(sub1, "file1.txt"))  # 文件被删除
    assert os.path.exists(os.path.join(sub2, "file2.txt"))  # 文件保留
    
    # 清理整个目录
    clear_cache(test_dir)
    
    # 检查目录重新创建
    assert os.path.exists(test_dir)
    assert not os.path.exists(sub2)


def test_disk_cache_with_custom_ttl():
    """测试 disk_cache 装饰器的动态 TTL 功能"""
    @disk_cache(path=temp_path, suffix="pkl")
    def func_with_ttl(x, **kwargs):
        return x * 3
    
    # 调用函数并设置短TTL
    result1 = func_with_ttl(10, ttl=1)
    assert result1 == 30
    
    # 立即再次调用应该使用缓存
    result2 = func_with_ttl(10, ttl=1)
    assert result2 == 30
    
    # 等待TTL过期
    time.sleep(2)
    
    # 再次调用应该重新计算
    result3 = func_with_ttl(10, ttl=1)
    assert result3 == 30


def test_disk_cache_error_handling():
    """测试 DiskCache 的错误处理"""
    cache = DiskCache()
    
    # 测试不支持的后缀
    try:
        cache.set("test", "data", suffix="unsupported")
        assert False, "应该抛出异常"
    except ValueError as e:
        assert "not supported" in str(e)
    
    # 测试 JSON 格式必须是字典
    try:
        cache.set("test", "not_dict", suffix="json")
        assert False, "应该抛出异常"
    except ValueError as e:
        assert "only support dict" in str(e)
    
    # 测试文本格式必须是字符串
    try:
        cache.set("test", 123, suffix="txt")
        assert False, "应该抛出异常"
    except ValueError as e:
        assert "only support str" in str(e)
    
    # 测试 DataFrame 格式
    try:
        cache.set("test", "not_dataframe", suffix="csv")
        assert False, "应该抛出异常"
    except ValueError as e:
        assert "only support pd.DataFrame" in str(e)


def test_disk_cache_get_nonexistent():
    """测试获取不存在的缓存文件"""
    cache = DiskCache()
    
    result = cache.get("nonexistent_key")
    assert result is None



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/9/24 12:39
describe: K线的笔特征计算
"""
import pandas as pd
from tqdm import tqdm
from typing import List
from czsc.objects import RawBar
from czsc.analyze import CZSC


def calculate_bi_info(bars: List[RawBar], **kwargs) -> pd.DataFrame:
    """计算笔的特征

    :param bars: 原始K线数据
    :return: 笔的特征
    """
    c = CZSC(bars, max_bi_num=kwargs.get("max_bi_num", 10000))

    res = [
        {
            "symbol": c.symbol,
            "sdt": bi.fx_a.dt,
            "edt": bi.fx_b.dt,
            "方向": bi.direction.value,
            "长度": bi.length,
            "分型数": len(bi.fxs),
            "斜边长度": bi.hypotenuse,
            "斜边角度": bi.angle,
            "涨跌幅": (bi.fx_b.fx / bi.fx_a.fx - 1) * 10000,
            "R2": bi.rsq,
        } for bi in c.bi_list
    ]
    _df = pd.DataFrame(res)
    _df['未来第一笔涨跌幅'] = _df['涨跌幅'].shift(-1)
    _df['未来第二笔涨跌幅'] = _df['涨跌幅'].shift(-2)
    return _df


def symbols_bi_infos(symbols, read_bars, freq='5分钟', sdt='20130101', edt='20190101', **kwargs) -> pd.DataFrame:
    """计算多个标的的笔特征

    :param symbols: 品种代码列表
    :param read_bars: 读取K线数据的函数，要求返回 RawBar 对象列表
    :param freq: K线周期, defaults to '5分钟'
    :param sdt: 开始时间, defaults to '20130101'
    :param edt: 结束时间, defaults to '20190101'
    :return: 笔的特征
    """
    bis = []
    for symbol in tqdm(symbols, desc="计算笔的特征"):
        try:
            bars = read_bars(symbol=symbol, freq=freq, sdt=sdt, edt=edt, fq='后复权')
            dfr = calculate_bi_info(bars)
            bis.append(dfr)
        except Exception as e:
            print(f"{symbol} 计算失败: {e}")
    dfb = pd.concat(bis, ignore_index=True)
    return dfb



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/6/6 13:56
describe: 阿里云OSS操作类
"""
import os
from loguru import logger
from tqdm import tqdm
from concurrent.futures import ThreadPoolExecutor
from io import BytesIO
from typing import List


class AliyunOSS:
    def __init__(self, access_key_id: str, access_key_secret: str, endpoint: str, bucket_name: str):
        """
        初始化AliyunOSS客户端。

        :param access_key_id: string, 阿里云账号AccessKey的ID。
        :param access_key_secret: string, 阿里云账号AccessKey的Secret。
        :param endpoint: string, OSS服务所在的区域的域名信息。
        :param bucket_name: string, 需要操作的Bucket的名称。
        """
        import oss2

        self.auth = oss2.Auth(access_key_id, access_key_secret)
        self.bucket = oss2.Bucket(self.auth, endpoint, bucket_name)

    def upload(self, filepath: str, oss_key: str, replace: bool = False) -> bool:
        """
        上传文件到OSS。

        :param filepath: string, 本地文件的路径。
        :param oss_key: string, 文件在OSS上的路径和名称。
        :param replace: boolean, 如果为True，将覆盖OSS上的同名文件。默认为False。
        :return: boolean, 如果上传成功，返回True；否则，返回False。
        """
        oss_key = oss_key.replace("\\", "/")
        dir_name = os.path.dirname(oss_key)
        if dir_name:
            self.create_folder(dir_name)

        if not replace and self.file_exists(oss_key):
            logger.info(f"{oss_key} exists in the bucket. Set replace=True to overwrite.")
            return False

        with open(filepath, "rb") as file:
            result = self.bucket.put_object(oss_key, file)
            if result.status == 200:
                logger.info(f"Upload {filepath} to {oss_key} successfully.")
                return True
            else:
                logger.error(f"Upload {filepath} to {oss_key} failed: {result.status}, {result.request_id}")
                return False

    def download(self, oss_key: str, filepath: str, replace: bool = False) -> bool:
        """
        从OSS下载文件。

        :param oss_key: string, 文件在OSS上的路径和名称。
        :param filepath: string, 本地存储文件的路径。
        :return: boolean, 如果下载成功，返回True；否则，返回False。
        """
        path = os.path.dirname(filepath)
        if not os.path.exists(path):
            os.makedirs(path)

        if not replace and os.path.exists(filepath):
            logger.info(f"{filepath} exists in the local. Set replace=True to overwrite.")
            return False

        result = self.bucket.get_object_to_file(oss_key, filepath)
        if result.status == 200:
            logger.info(f"Download {oss_key} to {filepath} successfully.")
            return True
        else:
            logger.error(f"Download {oss_key} to {filepath} failed: {result.status}, {result.request_id}")
            return False

    def delete_file(self, oss_key: str) -> bool:
        """
        从OSS删除文件。

        :param oss_key: string, 文件在OSS上的路径和名称。
        :return: boolean, 如果删除成功，返回True；否则，返回False.
        """
        result = self.bucket.delete_object(oss_key)
        if result.status == 204:
            logger.info(f"Delete {oss_key} successfully.")
            return True
        else:
            logger.error(f"Delete {oss_key} failed: {result.status}, {result.request_id}")
            return False

    def file_exists(self, oss_key: str) -> bool:
        """
        检查文件是否在OSS上存在。

        :param oss_key: string, 文件在OSS上的路径和名称。
        :return: boolean, 如果文件存在，返回True；否则，返回False.
        """
        return self.bucket.object_exists(oss_key)

    def get_file_stream(self, oss_key: str) -> BytesIO:
        """
        获取OSS上文件的数据流。

        :param oss_key: string, 文件在OSS上的路径和名称。
        :return: BytesIO, 文件的数据流。
        """
        result = self.bucket.get_object(oss_key)
        return BytesIO(result.read()) # type: ignore

    def create_folder(self, folder_path: str):
        """
        在OSS上创建文件夹。

        :param folder_path: string, 需要创建的文件夹的路径。
        """
        self.bucket.put_object(folder_path + "/", "")

    def list_files(self, prefix="", extensions=None):
        """
        列举OSS上的文件。

        :param prefix: string, 列举的文件前缀，默认为空。
        :param extensions: list, 需要列举的文件的后缀名，默认为空，表示列举所有文件。
        :return: list, 列举的文件的名称列表。
        """
        from oss2 import ObjectIterator

        oss_keys = []
        for obj in tqdm(ObjectIterator(self.bucket, prefix=prefix), desc=f"List files of {prefix}"):
            if obj.key.endswith("/"):
                continue
            if extensions and not any(obj.key.endswith(ext) for ext in extensions):
                continue
            oss_keys.append(obj.key)
        return oss_keys

    def batch_upload(self, filepaths: List[str], oss_keys: List[str], replace: bool = False, threads: int = 5):
        """
        批量上传文件到OSS。

        :param filepaths: list, 本地文件的路径列表。
        :param oss_keys: list, 文件在OSS上的路径和名称列表。
        :param replace: boolean, 如果为True，将覆盖OSS上的同名文件。默认为False。
        :param threads: int, 并行上传的线程数。默认为5。
        """
        with ThreadPoolExecutor(max_workers=threads) as executor:
            for filepath, oss_key in tqdm(zip(filepaths, oss_keys), total=len(filepaths), desc="Uploading"):
                executor.submit(self.upload, filepath, oss_key, replace)

    def batch_download(self, oss_keys: List[str], local_paths: List[str], replace: bool = False, threads: int = 5):
        """
        批量从OSS下载文件。

        :param oss_keys: list, 文件在OSS上的路径和名称列表。
        :param local_paths: list, 本地存储文件的路径列表。
        :param threads: int, 并行下载的线程数。默认为5。
        """
        assert len(oss_keys) == len(local_paths), "The length of oss_keys and local_paths must be the same."
        with ThreadPoolExecutor(max_workers=threads) as executor:
            for oss_key, local_path in zip(oss_keys, local_paths):
                executor.submit(self.download, oss_key, local_path, replace)

    def multipart_upload(self, filepath: str, oss_key: str):
        """
        分块上传大文件到OSS。

        :param filepath: string, 本地文件的路径。
        :param oss_key: string, 文件在OSS上的路径和名称。
        """
        from oss2.models import PartInfo
        from oss2 import determine_part_size

        total_size = os.path.getsize(filepath)
        part_size = determine_part_size(total_size, preferred_size=100 * 1024)
        upload_id = self.bucket.init_multipart_upload(oss_key).upload_id
        parts = []
        with open(filepath, "rb") as file:
            part_number = 1
            offset = 0
            while offset < total_size:
                num_to_upload = min(part_size, total_size - offset)
                result = self.bucket.upload_part(oss_key, upload_id, part_number, file.read(num_to_upload))
                parts.append(PartInfo(part_number, result.etag))
                offset += num_to_upload
                part_number += 1
        self.bucket.complete_multipart_upload(oss_key, upload_id, parts)

    def download_folder(self, oss_folder: str, local_folder: str, threads: int = 5):
        """
        从OSS下载指定文件夹。

        :param oss_folder: string, OSS上的文件夹路径。
        :param local_folder: string, 本地存储文件夹的路径。
        :param threads: int, 并行下载的线程数。默认为5。
        """
        if not os.path.exists(local_folder):
            os.makedirs(local_folder)

        oss_keys = self.list_files(prefix=oss_folder)
        file_paths = []
        for oss_key in oss_keys:
            file_path = os.path.join(local_folder, oss_key.replace(oss_folder, "", 1).lstrip("/\\"))
            file_paths.append(file_path)
        self.batch_download(oss_keys, file_paths, threads)

    def upload_folder(self, local_folder: str, oss_folder: str, replace: bool = False, threads: int = 5):
        """
        上传本地文件夹到OSS。

        :param local_folder: string, 本地文件夹的路径。
        :param oss_folder: string, OSS上的目标文件夹路径。
        :param replace: boolean, 如果为True，将覆盖OSS上的同名文件。默认为False。
        :param threads: int, 并行上传的线程数。默认为5。
        """
        filepaths = []
        oss_keys = []
        for root, dirs, files in os.walk(local_folder):
            for filename in files:
                filepath = os.path.join(root, filename)
                filepaths.append(filepath)

                relative_path = os.path.relpath(filepath, local_folder)
                oss_key = os.path.join(oss_folder, relative_path)
                oss_keys.append(oss_key)

        logger.info(f"Uploading {local_folder} to {oss_folder}, {len(filepaths)} files in total.")
        self.batch_upload(filepaths, oss_keys, replace, threads)
        logger.info(f"Upload {local_folder} finished.")



"""
价格敏感性分析模块

用于分析策略对执行价格的敏感性，通过对比不同交易价格的回测结果来评估价格执行对策略性能的影响。

主要功能：
1. 累计收益曲线展示
2. 价格敏感性分析
3. 分析结果摘要

作者: czsc
"""

import pandas as pd
import streamlit as st
from loguru import logger
from typing import Optional, Tuple

from .base import safe_import_weight_backtest, apply_stats_style
from .returns import show_cumulative_returns


def show_price_sensitive(df: pd.DataFrame, 
                        fee: float = 2.0, 
                        digits: int = 2, 
                        weight_type: str = "ts", 
                        n_jobs: int = 1, 
                        **kwargs) -> Optional[Tuple[pd.DataFrame, pd.DataFrame]]:
    """价格敏感性分析组件
    
    分析策略对执行价格的敏感性，通过对比不同交易价格的回测结果来评估价格执行对策略性能的影响。
    
    参数:
        df: 包含以下必要列的数据框：
            - symbol: 合约代码
            - dt: 日期时间
            - weight: 仓位权重
            - price: 基准价格
            - TP*: 以TP开头的交易价格列（如TP_open, TP_high等）
        fee: 单边费率（BP），默认2.0
        digits: 小数位数，默认2
        weight_type: 权重类型，可选 "ts" 或 "cs"，默认 "ts"
        n_jobs: 并行数，默认1
        **kwargs: 其他参数
            - title_prefix: 标题前缀，默认为空
            - show_detailed_stats: 是否显示详细统计信息，默认False
    
    返回:
        tuple: (dfr, dfd) 分别为统计结果DataFrame和日收益率DataFrame，失败时返回None
    
    示例:
        >>> # 基本用法
        >>> dfr, dfd = show_price_sensitive(df, fee=2.0, digits=2)
        
        >>> # 自定义参数
        >>> dfr, dfd = show_price_sensitive(
        ...     df, 
        ...     fee=1.5, 
        ...     digits=3, 
        ...     weight_type="cs",
        ...     n_jobs=4,
        ...     title_prefix="策略A - ",
        ...     show_detailed_stats=True
        ... )
    """
    from czsc.eda import cal_yearly_days
    
    # 参数处理
    title_prefix = kwargs.get("title_prefix", "")
    show_detailed_stats = kwargs.get("show_detailed_stats", False)
    
    # 数据验证
    required_cols = ["symbol", "dt", "weight", "price"]
    missing_cols = [col for col in required_cols if col not in df.columns]
    if missing_cols:
        error_msg = f"缺少必要的列: {missing_cols}"
        st.error(error_msg)
        logger.error(f"数据检查失败，{error_msg}")
        return None
    
    # 查找交易价格列
    tp_cols = [x for x in df.columns if x.startswith("TP")]
    if not tp_cols:
        error_msg = "没有找到交易价格列，请检查文件；交易价列名必须以 TP 开头"
        st.error(error_msg)
        logger.error("未找到以TP开头的交易价格列")
        return None
    
    logger.info(f"找到 {len(tp_cols)} 个交易价格列: {tp_cols}")
    
    # 计算年化天数
    try:
        yearly_days = cal_yearly_days(dts=df["dt"].unique().tolist())
        logger.info(f"计算得到年化天数: {yearly_days}")
    except Exception as e:
        error_msg = f"计算年化天数失败: {e}"
        st.error(error_msg)
        logger.error(error_msg)
        return None
    
    # 获取WeightBacktest类
    WeightBacktest = safe_import_weight_backtest()
    if WeightBacktest is None:
        st.error("无法导入回测类，请检查 czsc 安装")
        return None
    
    # 结果收集
    c1 = st.container(border=True)
    rows = []
    dfd = pd.DataFrame()
    
    # 创建进度条
    progress_bar = st.progress(0)
    status_text = st.empty()
    
    # 逐个处理交易价格
    for i, tp_col in enumerate(tp_cols):
        try:
            progress = (i + 1) / len(tp_cols)
            progress_bar.progress(progress)
            status_text.text(f"正在处理第 {i+1}/{len(tp_cols)} 个交易价格: {tp_col}")
            
            logger.info(f"正在处理第 {i+1}/{len(tp_cols)} 个交易价格: {tp_col}")
            
            # 准备数据
            df_temp = df.copy()
            df_temp[tp_col] = df_temp[tp_col].fillna(df_temp["price"])
            dfw = df_temp[["symbol", "dt", 'weight', tp_col]].copy()
            dfw.rename(columns={tp_col: "price"}, inplace=True)

            # 创建回测实例
            wb = WeightBacktest(
                dfw=dfw,
                digits=digits,
                fee_rate=fee/10000,
                weight_type=weight_type,
                n_jobs=n_jobs,
                yearly_days=yearly_days
            )
            
            # 获取日收益率
            daily = wb.daily_return.copy()
            daily.rename(columns={"total": tp_col}, inplace=True)
            
            if dfd.empty:
                dfd = daily[['date', tp_col]].copy()
            else:
                dfd = pd.merge(dfd, daily[['date', tp_col]], on='date', how='outer')

            # 收集统计结果
            res = {"交易价格": tp_col}
            res.update(wb.stats)
            rows.append(res)
            
        except Exception as e:
            warning_msg = f"处理交易价格 {tp_col} 时出错: {e}"
            st.warning(warning_msg)
            logger.error(warning_msg)
            continue
    
    # 清除进度条
    progress_bar.empty()
    status_text.empty()
    
    if not rows:
        st.error("所有交易价格处理失败，无法生成报告")
        return None


    # 显示结果
    with c1:
        st.markdown(f"##### :red[{title_prefix}不同交易价格回测核心指标对比]")
        dfr = pd.DataFrame(rows)

        # 敏感性评估
        if len(dfr) > 1 and "年化" in dfr.columns:
            _show_sensitivity_assessment(dfr)

        # 选择显示列
        if show_detailed_stats:
            display_cols = ['交易价格', "开始日期", "结束日期", "绝对收益", "年化", "年化波动率", 
                           "夏普", "最大回撤", "卡玛", "日胜率", "日盈亏比", "交易胜率", 
                           "单笔收益", "持仓K线数", "持仓天数", "多头占比", "空头占比"]
        else:
            display_cols = ['交易价格', "开始日期", "结束日期", "绝对收益", "年化", "年化波动率", 
                           "夏普", "最大回撤", "卡玛", "交易胜率", "单笔收益", "持仓K线数"]
        
        # 确保所有列都存在
        available_cols = [col for col in display_cols if col in dfr.columns]
        dfr_display = dfr[available_cols].copy()
        
        # 应用样式
        dfr_styled = apply_stats_style(dfr_display)
        st.dataframe(dfr_styled, use_container_width=True)
    
    # 累计收益对比
    c2 = st.container(border=True)
    with c2:
        st.markdown(f"##### :red[{title_prefix}不同交易价格回测累计收益对比]")
        
        if not dfd.empty:
            dfd_plot = dfd.copy()
            dfd_plot['date'] = pd.to_datetime(dfd_plot['date'])
            dfd_plot.set_index("date", inplace=True)
            
            show_cumulative_returns(
                dfd_plot, 
                fig_title=f"{title_prefix}不同交易价格累计收益对比"
            )
        else:
            st.warning("没有有效的收益率数据用于绘制图表")
    
    logger.info(f"价格敏感性分析完成，共处理 {len(rows)} 个交易价格")
    return dfr, dfd


def _show_sensitivity_assessment(dfr: pd.DataFrame) -> None:
    """显示敏感性评估"""
    annual_returns = dfr["年化"]
    sensitivity_score = (annual_returns.max() - annual_returns.min()) / annual_returns.mean()
    
    st.markdown("**敏感性评估：**")
    if sensitivity_score < 0.1:
        st.success(f"🟢 策略对价格执行不敏感 (敏感度: {sensitivity_score:.2%})")
    elif sensitivity_score < 0.3:
        st.warning(f"🟡 策略对价格执行中等敏感 (敏感度: {sensitivity_score:.2%})")
    else:
        st.error(f"🔴 策略对价格执行高度敏感 (敏感度: {sensitivity_score:.2%})")


# 支持的函数列表
__all__ = [
    'show_price_sensitive', 
] 


# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/4/14 19:27
describe:
"""
from czsc.analyze import CZSC
from collections import OrderedDict
from czsc.traders.base import CzscTrader
from czsc.utils import create_single_signal, get_sub_elements
from czsc.objects import Operate, Direction, Mark
from czsc.signals.tas import update_ma_cache


def pos_ma_V230414(cat: CzscTrader, **kwargs) -> OrderedDict:
    """判断开仓后是否升破MA均线或跌破MA均线

    参数模板："{pos_name}_{freq1}#{ma_type}#{timeperiod}_持有状态V230414"

    **信号逻辑：**

    多头持有状态如下，反之为空头持有状态：

    1. 如果持有多头，且开仓后有价格升破MA均线，则为多头升破均线；
    2. 如果持有空头，且开仓后有价格跌破MA均线，则为空头跌破均线。

    **信号列表：**

    - Signal('日线三买多头N1_60分钟#SMA#5_持有状态V230414_多头_升破均线_任意_0')
    - Signal('日线三买多头N1_60分钟#SMA#5_持有状态V230414_空头_跌破均线_任意_0')

    :param cat: CzscTrader对象
    :param kwargs: 参数字典
        - pos_name: str，开仓信号的名称
        - freq1: str，给定的K线周期
        - n: int，向前找的分型个数，默认为 3
    :return:
    """
    pos_name = kwargs["pos_name"]
    freq1 = kwargs["freq1"]
    ma_type = kwargs.get("ma_type", "SMA").upper()
    timeperiod = int(kwargs.get("timeperiod", 5))
    k1, k2, k3 = f"{pos_name}_{freq1}#{ma_type}#{timeperiod}_持有状态V230414".split("_")
    v1, v2 = "其他", "其他"
    key = update_ma_cache(cat.kas[freq1], ma_type=ma_type, timeperiod=timeperiod)
    # 如果没有持仓策略，则不产生信号
    if not hasattr(cat, "positions"):
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    pos = [x for x in cat.positions if x.name == pos_name][0]
    if len(pos.operates) == 0 or pos.operates[-1]["op"] in [Operate.SE, Operate.LE]:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    c = cat.kas[freq1]
    op = pos.operates[-1]

    if op["op"] == Operate.LO:
        bars = [x for x in c.bars_raw[-100:] if x.dt > op["dt"]]
        for x in bars:
            if x.close > x.cache[key]:
                v1, v2 = "多头", "升破均线"
                break

    if op["op"] == Operate.SO:
        bars = [x for x in c.bars_raw[-100:] if x.dt > op["dt"]]
        for x in bars:
            if x.close < x.cache[key]:
                v1, v2 = "空头", "跌破均线"
                break

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1, v2=v2)


def pos_fx_stop_V230414(cat: CzscTrader, **kwargs) -> OrderedDict:
    """按照开仓点附近的分型止损

    参数模板："{freq1}_{pos_name}N{n}_止损V230414"

    **信号逻辑：**

    多头止损逻辑如下，反之为空头止损逻辑：

    1. 从多头开仓点开始，在给定对的K线周期 freq1 上向前找 N 个底分型，记为 F1
    2. 将这 N 个底分型的最低点，记为 L1，如果最新价低于 L1，则止损

    **信号列表：**

    - Signal('日线_日线三买多头N1_止损V230414_多头止损_任意_任意_0')
    - Signal('日线_日线三买多头N1_止损V230414_空头止损_任意_任意_0')

    :param cat: CzscTrader对象
    :param kwargs: 参数字典
        - pos_name: str，开仓信号的名称
        - freq1: str，给定的K线周期
        - n: int，向前找的分型个数，默认为 3
    :return:
    """
    pos_name = kwargs["pos_name"]
    freq1 = kwargs["freq1"]
    n = int(kwargs.get("n", 3))
    k1, k2, k3 = f"{freq1}_{pos_name}N{n}_止损V230414".split("_")
    v1 = "其他"

    # 如果没有持仓策略，则不产生信号
    if not hasattr(cat, "positions"):
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    pos = [x for x in cat.positions if x.name == pos_name][0]
    if len(pos.operates) == 0 or pos.operates[-1]["op"] in [Operate.SE, Operate.LE]:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    c = cat.kas[freq1]
    op = pos.operates[-1]

    # 多头止损逻辑
    if op["op"] == Operate.LO:
        fxs = [x for x in c.fx_list if x.mark == Mark.D and x.dt < op["dt"]][-n:]
        if cat.latest_price < min([x.low for x in fxs]):
            v1 = "多头止损"

    # 空头止损逻辑
    if op["op"] == Operate.SO:
        fxs = [x for x in c.fx_list if x.mark == Mark.G and x.dt < op["dt"]][-n:]
        if cat.latest_price > max([x.high for x in fxs]):
            v1 = "空头止损"

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def pos_bar_stop_V230524(cat: CzscTrader, **kwargs) -> OrderedDict:
    """按照开仓点附近的N根K线极值止损

    参数模板："{pos_name}_{freq1}N{n}K_止损V230524"

    **信号逻辑：**

    多头止损逻辑如下，反之为空头止损逻辑：

    1. 从多头开仓点开始，在给定的K线周期 freq1 上向前找 N 个K线，记为 F1
    2. 将这 N 个K线的最低点，记为 L1，如果最新价跌破 L1，则止损

    **信号列表：**

    - Signal('日线三买多头_日线N3K_止损V230524_多头止损_任意_任意_0')
    - Signal('日线三买多头_日线N3K_止损V230524_空头止损_任意_任意_0')

    :param cat: CzscTrader对象
    :param kwargs: 参数字典

        - pos_name: str，开仓信号的名称
        - freq1: str，给定的K线周期
        - n: int，向前找的K线个数，默认为 3

    :return: OrderedDict
    """
    pos_name = kwargs["pos_name"]
    freq1 = kwargs["freq1"]
    n = int(kwargs.get("n", 3))
    k1, k2, k3 = f"{pos_name}_{freq1}N{n}K_止损V230524".split("_")
    v1 = "其他"
    assert 20 >= n >= 1, "参数 n 取值范围为 1~20"
    # 如果没有持仓策略，则不产生信号
    if not hasattr(cat, "positions"):
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    pos_ = [x for x in cat.positions if x.name == pos_name][0]
    if len(pos_.operates) == 0 or pos_.operates[-1]["op"] in [Operate.SE, Operate.LE]:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    c: CZSC = cat.kas[freq1]
    op = pos_.operates[-1]

    # 多头止损逻辑
    if op["op"] == Operate.LO:
        bars = [x for x in c.bars_raw[-100:] if x.dt < op["dt"]][-n:]
        if cat.latest_price < min([x.low for x in bars]):
            v1 = "多头止损"

    # 空头止损逻辑
    if op["op"] == Operate.SO:
        bars = [x for x in c.bars_raw[-100:] if x.dt < op["dt"]][-n:]
        if cat.latest_price > max([x.high for x in bars]):
            v1 = "空头止损"

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def pos_holds_V230414(cat: CzscTrader, **kwargs) -> OrderedDict:
    """开仓后N根K线涨幅小于M%%，则平仓

    参数模板："{pos_name}_{freq1}N{n}M{m}_趋势判断V230414"

    **信号逻辑：**

    1. 找出开仓后的 N 根K线，计算涨幅，如果涨幅小于 M%%，则平仓
    2. 这里面的逻辑是，如果开仓后的 N 根K线涨幅小于 M%%，则说明趋势不明朗，平仓等待

    **信号列表：**

    - Signal('日线三买多头N1_60分钟N5M100_趋势判断V230414_多头存疑_任意_任意_0')
    - Signal('日线三买多头N1_60分钟N5M100_趋势判断V230414_多头良好_任意_任意_0')

    :param cat: CzscTrader对象
    :param kwargs: 参数字典
        - pos_name: str，开仓信号的名称
        - freq1: str，给定的K线周期
        - n: int，最少持有K线数量，默认为 5，表示5根K线之后开始判断趋势
        - m: int，涨幅阈值，默认为 100，表示涨幅小于 100BP 时，平仓
    :return:
    """
    pos_name = kwargs["pos_name"]
    freq1 = kwargs["freq1"]
    n = int(kwargs.get("n", 5))
    m = int(kwargs.get("m", 100))
    k1, k2, k3 = f"{pos_name}_{freq1}N{n}M{m}_趋势判断V230414".split("_")
    v1 = "其他"
    # 如果没有持仓策略，则不产生信号
    if not hasattr(cat, "positions"):
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    pos = [x for x in cat.positions if x.name == pos_name][0]
    if len(pos.operates) == 0 or pos.operates[-1]["op"] in [Operate.SE, Operate.LE]:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    c = cat.kas[freq1]
    op = pos.operates[-1]
    bars = [x for x in c.bars_raw[-100:] if x.dt > op["dt"]]
    if len(bars) < n:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    if op["op"] == Operate.LO:
        zdf = (bars[-1].close - op["price"]) / op["price"] * 10000
        v1 = "多头存疑" if zdf < m else "多头良好"

    if op["op"] == Operate.SO:
        zdf = (op["price"] - bars[-1].close) / op["price"] * 10000
        v1 = "空头存疑" if zdf < m else "空头良好"

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def pos_fix_exit_V230624(cat: CzscTrader, **kwargs) -> OrderedDict:
    """固定比例止损，止盈

    参数模板："{pos_name}_固定{th}BP止盈止损_出场V230624"

    **信号逻辑：**

    以多头为例，如果持有收益超过 th 个BP，则止盈；如果亏损超过 th 个BP，则止损。

    **信号列表：**

    - Signal('日线三买多头_固定100BP止盈止损_出场V230624_多头止损_任意_任意_0')
    - Signal('日线三买多头_固定100BP止盈止损_出场V230624_空头止损_任意_任意_0')

    :param cat: CzscTrader对象
    :param kwargs: 参数字典
        - pos_name: str，开仓信号的名称
        - freq1: str，给定的K线周期
        - n: int，向前找的K线个数，默认为 3
    :return:
    """
    pos_name = kwargs["pos_name"]
    th = int(kwargs.get("th", 300))
    k1, k2, k3 = f"{pos_name}_固定{th}BP止盈止损_出场V230624".split("_")
    v1 = "其他"
    if not hasattr(cat, "positions"):
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    pos_ = [x for x in cat.positions if x.name == pos_name][0]
    if len(pos_.operates) == 0 or pos_.operates[-1]["op"] in [Operate.SE, Operate.LE]:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    op = pos_.operates[-1]
    op_price = op["price"]

    if op["op"] == Operate.LO:
        if cat.latest_price < op_price * (1 - th / 10000):
            v1 = "多头止损"
        if cat.latest_price > op_price * (1 + th / 10000):
            v1 = "多头止盈"

    if op["op"] == Operate.SO:
        if cat.latest_price > op_price * (1 + th / 10000):
            v1 = "空头止损"
        if cat.latest_price < op_price * (1 - th / 10000):
            v1 = "空头止盈"

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def pos_profit_loss_V230624(cat: CzscTrader, **kwargs) -> OrderedDict:
    """开仓后盈亏比达到一定比值，才允许平仓  贡献者：谌意勇

    参数模板："{pos_name}_{freq1}YKB{ykb}N{n}_盈亏比判断V230624"

    **信号逻辑：**

    1. 通过公式 计算盈亏比=abs(现价-开仓价）/abs(开仓价-止损价）* 10,当比值大于一定阀值时才允许平仓

    **信号列表：**

    - Signal('日线通道突破_60分钟YKB20N3_盈亏比判断V230624_空头止损_任意_任意_0')
    - Signal('日线通道突破_60分钟YKB20N3_盈亏比判断V230624_多头止损_任意_任意_0')
    - Signal('日线通道突破_60分钟YKB20N3_盈亏比判断V230624_多头达标_任意_任意_0')
    - Signal('日线通道突破_60分钟YKB20N3_盈亏比判断V230624_空头达标_任意_任意_0')

    :param cat: CzscTrader对象
    :param kwargs: 参数字典

        - pos_name: str，开仓信号的名称
        - freq1: str，给定的K线周期
        - ykb: int，默认为 20, 表示2倍盈亏比，计算盈亏比=abs(现价-开仓价）/abs(开仓价-止损价）
        - n: int 默认为3  止损取最近n个分型的最低点或最高点

    :return:
    """
    pos_name = kwargs["pos_name"]
    freq1 = kwargs["freq1"]
    ykb = int(kwargs.get("ykb", 20))
    n = int(kwargs.get("n", 3))
    k1, k2, k3 = f"{pos_name}_{freq1}YKB{ykb}N{n}_盈亏比判断V230624".split("_")
    v1 = "其他"
    # 如果没有持仓策略，则不产生信号
    if not hasattr(cat, "positions"):
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    pos = [x for x in cat.positions if x.name == pos_name][0]
    if len(pos.operates) == 0 or pos.operates[-1]["op"] in [Operate.SE, Operate.LE]:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    c = cat.kas[freq1]
    op = pos.operates[-1]
    last_close = c.bars_raw[-1].close

    if op["op"] == Operate.LO:
        fxs = [x for x in c.fx_list if x.mark == Mark.D and x.dt < op["dt"]][-n:]
        stop_price = min([x.low for x in fxs])
        ykb_ = ((last_close - op["price"]) / (op["price"] - stop_price)) * 10
        if ykb_ > ykb:
            v1 = "多头达标"
        else:
            if last_close < stop_price:
                v1 = "多头止损"

    if op["op"] == Operate.SO:
        fxs = [x for x in c.fx_list if x.mark == Mark.G and x.dt < op["dt"]][-n:]
        stop_price = max([x.high for x in fxs])
        ykb_ = ((last_close - op["price"]) / (op["price"] - stop_price)) * 10
        if ykb_ > ykb:
            v1 = "空头达标"
        else:
            if last_close > stop_price:
                v1 = "空头止损"

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def pos_status_V230808(cat: CzscTrader, **kwargs) -> OrderedDict:
    """Position策略的持仓状态

    参数模板："{pos_name}_持仓状态_BS辅助V230808"

    **信号逻辑：**

    对指定的持仓策略，有三种状态：持多、持空、持币。

    **信号列表：**

    - Signal('日线三买多头N1_持仓状态_BS辅助V230808_持多_任意_任意_0')
    - Signal('日线三买多头N1_持仓状态_BS辅助V230808_持空_任意_任意_0')
    - Signal('日线三买多头N1_持仓状态_BS辅助V230808_持币_任意_任意_0')

    :param cat: CzscTrader对象
    :param kwargs: 参数字典
        - pos_name: str，开仓信号的名称
    :return:
    """
    pos_name = kwargs["pos_name"]
    k1, k2, k3 = f"{pos_name}_持仓状态_BS辅助V230808".split("_")
    # 如果没有持仓策略，则不产生信号
    if not hasattr(cat, "positions"):
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1="其他")

    pos = cat.get_position(pos_name)
    v1 = "持币"
    if len(pos.operates) == 0:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    op = pos.operates[-1]
    if op["op"] == Operate.LO:
        v1 = "持多"
    if op["op"] == Operate.SO:
        v1 = "持空"
    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def pos_holds_V230807(cat: CzscTrader, **kwargs) -> OrderedDict:
    """开仓后N根K线收益小于M%%，且当前收益大于T%%，平仓保本

    参数模板："{pos_name}_{freq1}N{n}M{m}T{t}_BS辅助V230807"

    **信号逻辑：**

    1. 针对某个开仓点，如果 N 根K线之后，收益低于 M，则认为开仓失误；
    2. 开仓的失误发生后，如果市场给了逃命的机会，不贪心，等待收益大于 T 时，平仓保本；
    3. 保本有两种场景：开仓后先亏损后反弹；开仓后先盈利后回落。

    **信号列表：**

    - Signal('日线三买多头N1_60分钟N5M50T10_BS辅助V230807_多头保本_任意_任意_0')
    - Signal('日线三买多头N1_60分钟N5M50T10_BS辅助V230807_空头保本_任意_任意_0')

    :param cat: CzscTrader对象
    :param kwargs: 参数字典

        - pos_name: str，开仓信号的名称
        - freq1: str，给定的K线周期
        - n: int，最少持有K线数量，默认为 5，表示5根K线之后开始判断
        - m: int，收益最小阈值，默认为 100，表示收益小于 100BP 时，需要开始判断保本单
        - t: int，保本收益阈值，默认为 10，表示收益大于 10BP 时，可以平仓保本

    :return:
    """
    pos_name = kwargs["pos_name"]
    freq1 = kwargs["freq1"]
    n = int(kwargs.get("n", 5))
    m = int(kwargs.get("m", 50))
    t = int(kwargs.get("t", 10))
    assert m > t > 0, "参数 m 必须大于 t"
    k1, k2, k3 = f"{pos_name}_{freq1}N{n}M{m}T{t}_BS辅助V230807".split("_")
    v1 = "其他"
    # 如果没有持仓策略，则不产生信号
    if not cat.kas or not hasattr(cat, "positions"):
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    pos = [x for x in cat.positions if x.name == pos_name][0]
    if len(pos.operates) == 0 or pos.operates[-1]["op"] in [Operate.SE, Operate.LE]:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    c = cat.kas[freq1]
    op = pos.operates[-1]
    bars = [x for x in c.bars_raw[-100:] if x.dt > op["dt"]]
    if len(bars) < n:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    if op["op"] == Operate.LO:
        zdf = (bars[-1].close - op["price"]) / op["price"] * 10000
        if t < zdf < m:
            v1 = "多头保本"

    if op["op"] == Operate.SO:
        zdf = (op["price"] - bars[-1].close) / op["price"] * 10000
        if t < zdf < m:
            v1 = "空头保本"

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def pos_holds_V240428(cat: CzscTrader, **kwargs) -> OrderedDict:
    """保本单：开仓后最大盈利超过H个BP，且当前收益低于最大盈利的T%，平仓保本

    参数模板："{pos_name}_{freq1}H{h}T{t}N{n}_保本V240428"

    **信号逻辑：**

    以多头保本单为例，计算过程如下：

    1. 从多头开仓点开始，在给定的K线周期 freq1 上计算开仓后的最大盈利，记为 Y1；
    2. 计算当前收益，记为 Y2；
    3. 如果Y1 大于H，且 Y2 < Y1 * T / 100，则平仓保本。

    **信号列表：**

    - Signal('日线三买多头N1_60分钟H100T20N5_保本V240428_空头保本_任意_任意_0')
    - Signal('日线三买多头N1_60分钟H100T20N5_保本V240428_多头保本_任意_任意_0')

    :param cat: CzscTrader对象
    :param kwargs: 参数字典

        - pos_name: str，开仓信号的名称
        - freq1: str，给定的K线周期
        - h: int，最大盈利，单位BP，默认为 100
        - t: int，最大盈利的T%，默认为 20
        - n: int，最少持有K线数量，默认为 5，表示5根K线之后开始判断

    :return: OrderedDict
    """
    pos_name = kwargs["pos_name"]
    freq1 = kwargs["freq1"]
    h = int(kwargs.get("h", 100))  # 最大盈利，单位BP
    t = int(kwargs.get("t", 20))  # 最大盈利的T%
    n = int(kwargs.get("n", 5))  # 最少持有K线数量，默认为 5，表示5根K线之后开始判断

    k1, k2, k3 = f"{pos_name}_{freq1}H{h}T{t}N{n}_保本V240428".split("_")
    v1 = "其他"

    # 如果没有持仓策略，则不产生信号
    if not cat.kas or not hasattr(cat, "positions"):
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    pos = [x for x in cat.positions if x.name == pos_name][0]
    if len(pos.operates) == 0 or pos.operates[-1]["op"] in [Operate.SE, Operate.LE]:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    c = cat.kas[freq1]
    op = pos.operates[-1]
    bars = [x for x in c.bars_raw[-100:] if x.dt > op["dt"]]
    if len(bars) < n:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    if op["op"] == Operate.LO:
        y1 = (max([x.close for x in bars]) - op["price"]) / op["price"] * 10000
        y2 = (bars[-1].close - op["price"]) / op["price"] * 10000
        if y1 > h and y2 < y1 * t / 100:
            v1 = "多头保本"

    if op["op"] == Operate.SO:
        y1 = (op["price"] - min([x.close for x in bars])) / op["price"] * 10000
        y2 = (op["price"] - bars[-1].close) / op["price"] * 10000
        if y1 > h and y2 < y1 * t / 100:
            v1 = "空头保本"

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def pos_holds_V240608(cat: CzscTrader, **kwargs) -> OrderedDict:
    """保本单：多头开仓后，最低价跌破前低，当前价在成本价上方N个价位，平仓保本；空头反之。

    参数模板："{pos_name}_{freq1}W{w}N{n}_保本V240608"

    **信号逻辑：**

    以多头保本单为例，计算过程如下：

    1. 从多头开仓点开始，在给定的K线周期 freq1 上计算开仓前 W 个K线的最低价，记为 L1；
    2. 计算开仓后的最低价，记为 L2；
    3. 如果 L2 < L1，且当前价比开仓价高 N 个价位，则平仓保本。

    **信号列表：**

    - Signal('日线三买多头N1_60分钟W20N2_保本V240608_空头保本_任意_任意_0')
    - Signal('日线三买多头N1_60分钟W20N2_保本V240608_多头保本_任意_任意_0')

    :param cat: CzscTrader对象
    :param kwargs: 参数字典

        - pos_name: str，开仓信号的名称
        - freq1: str，给定的K线周期
        - w: int，开仓前W根K线，默认为 20
        - n: int，成本价上方N个价位，默认为 2

    :return: OrderedDict
    """
    pos_name = kwargs["pos_name"]
    freq1 = kwargs["freq1"]
    w = int(kwargs.get("w", 20))  # 开仓前W根K线
    n = int(kwargs.get("n", 2))  # 成本价上方N个价位

    k1, k2, k3 = f"{pos_name}_{freq1}W{w}N{n}_保本V240608".split("_")
    v1 = "其他"

    # 如果没有持仓策略，则不产生信号
    if not cat.kas or not hasattr(cat, "positions"):
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    pos = [x for x in cat.positions if x.name == pos_name][0]
    if len(pos.operates) == 0 or pos.operates[-1]["op"] in [Operate.SE, Operate.LE]:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    c = cat.kas[freq1]
    op = pos.operates[-1]

    # 开仓前W根K线
    w_bars = [x for x in c.bars_raw[-200:] if x.dt <= op["dt"]][-w:]
    # 开仓后的K线
    a_bars = [x for x in c.bars_raw[-100:] if x.dt > op["dt"]]
    unique_prices = [p for x in c.bars_raw[-200:] for p in [x.high, x.low, x.close, x.open]]
    unique_prices = sorted(list(set(unique_prices)))

    if op["op"] == Operate.LO and w_bars:
        w_low = min([x.low for x in w_bars])  # 开仓前最低价
        a_low = min([x.low for x in a_bars])  # 开仓后最低价
        up_prices = [x for x in unique_prices if x > op["price"]]  # 成本价上方的价位
        # 如果开仓后的最低价低于开仓前的最低价，且当前价比开仓价高 N 个价位，则平仓保本
        if len(up_prices) > n and a_low < w_low and cat.latest_price > up_prices[n]:
            v1 = "多头保本"

    if op["op"] == Operate.SO and w_bars:
        w_high = max([x.high for x in w_bars])  # 开仓前最高价
        a_high = max([x.high for x in a_bars])  # 开仓后最高价
        down_prices = [x for x in unique_prices if x < op["price"]]  # 成本价下方的价位
        # 如果开仓后的最高价高于开仓前的最高价，且当前价比开仓价低 N 个价位，则平仓保本
        if len(down_prices) > n and a_high > w_high and cat.latest_price < down_prices[-n]:
            v1 = "空头保本"

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def pos_stop_V240428(cat: CzscTrader, **kwargs) -> OrderedDict:
    """止损单，持有N根K线后，多头跌破前低或空头升破前高，平仓

    参数模板："{pos_name}_{freq1}T{t}N{n}_止损V240428"

    **信号逻辑：**

    以多头止损为例，计算过程如下：

    1. 从多头开仓点开始，在给定的K线周期 freq1 上计算开仓 N 根K线后的最新价 close；
    2. 计算开仓前的 unique_price 列表，获取低于开仓价的列表，降序排列后的第 t 个价位作为止损价 Y；
    3. 如果 close < Y，则止损平仓。

    **信号列表：**

    - Signal('日线三买多头N1_60分钟T5N5_止损V240428_空头止损_任意_任意_0')
    - Signal('日线三买多头N1_60分钟T5N5_止损V240428_多头止损_任意_任意_0')

    :param cat: CzscTrader对象
    :param kwargs: 参数字典

        - pos_name: str，开仓信号的名称
        - freq1: str，给定的K线周期
        - t: int，止损多少跳，默认为 20
        - n: int，最少持有K线数量，默认为 5，表示5根K线之后开始判断

    :return: OrderedDict
    """
    pos_name = kwargs["pos_name"]
    freq1 = kwargs["freq1"]
    t = int(kwargs.get("t", 20))  # 止损多少跳
    n = int(kwargs.get("n", 5))  # 最少持有K线数量，默认为 5，表示5根K线之后开始判断

    k1, k2, k3 = f"{pos_name}_{freq1}T{t}N{n}_止损V240428".split("_")
    v1 = "其他"

    # 如果没有持仓策略，则不产生信号
    if not cat.kas or not hasattr(cat, "positions"):
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    pos = [x for x in cat.positions if x.name == pos_name][0]
    if len(pos.operates) == 0 or pos.operates[-1]["op"] in [Operate.SE, Operate.LE]:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    c = cat.kas[freq1]
    op = pos.operates[-1]
    right_bars = [x for x in c.bars_raw[-100:] if x.dt > op["dt"]]

    if len(right_bars) < n:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    left_bars = [x for x in c.bars_raw if x.dt < op["dt"]]
    unique_prices = [p for x in left_bars for p in [x.high, x.low, x.close, x.open]]
    unique_prices = sorted(list(set(unique_prices)))

    if op["op"] == Operate.LO:
        low_prices = sorted([x for x in unique_prices if x < op["price"]], reverse=True)
        if not low_prices:
            return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

        y = low_prices[-t] if len(low_prices) > t else low_prices[0]
        if right_bars[-1].close < y:
            v1 = "多头止损"

    if op["op"] == Operate.SO:
        high_prices = sorted([x for x in unique_prices if x > op["price"]])
        if not high_prices:
            return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

        y = high_prices[t] if len(high_prices) > t else high_prices[-1]
        if right_bars[-1].close > y:
            v1 = "空头止损"

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def pos_take_V240428(cat: CzscTrader, **kwargs) -> OrderedDict:
    """止盈单，持有N根K线后，多头持仓期间出现T根倍量阳线或空头持仓期间出现T根倍量阴线，平仓

    参数模板："{pos_name}_{freq1}T{t}N{n}_止盈V240428"

    **信号逻辑：**

    以多头为例，计算过程如下：

    1. 从多头开仓点后N根K线开始，寻找倍量阳线，计算数量为 C；
    2. 如果 C >= T，则止盈平仓。

    **信号列表：**

    - Signal('日线三买多头N1_60分钟T5N5_止盈V240428_空头止盈_任意_任意_0')
    - Signal('日线三买多头N1_60分钟T5N5_止盈V240428_多头止盈_任意_任意_0')

    :param cat: CzscTrader对象
    :param kwargs: 参数字典

        - pos_name: str，开仓信号的名称
        - freq1: str，给定的K线周期
        - t: int，倍量K线数量止盈，默认为 3
        - n: int，最少持有K线数量，默认为 5，表示5根K线之后开始判断

    :return: OrderedDict
    """
    pos_name = kwargs["pos_name"]
    freq1 = kwargs["freq1"]
    t = int(kwargs.get("t", 3))  # 倍量K线数量止盈
    n = int(kwargs.get("n", 5))  # 最少持有K线数量，默认为 5，表示5根K线之后开始判断

    k1, k2, k3 = f"{pos_name}_{freq1}T{t}N{n}_止盈V240428".split("_")
    v1 = "其他"

    # 如果没有持仓策略，则不产生信号
    if not cat.kas or not hasattr(cat, "positions"):
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    pos = [x for x in cat.positions if x.name == pos_name][0]
    if len(pos.operates) == 0 or pos.operates[-1]["op"] in [Operate.SE, Operate.LE]:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    c = cat.kas[freq1]
    op = pos.operates[-1]
    bars = [x for x in c.bars_raw[-100:] if x.dt > op["dt"]]

    if len(bars) < n:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    if op["op"] == Operate.LO:
        c1 = 0
        for i in range(1, len(bars)):
            if bars[i].close > bars[i].open and bars[i].vol > bars[i - 1].vol * 2:
                c1 += 1
        if c1 >= t:
            v1 = "多头止盈"

    if op["op"] == Operate.SO:
        c2 = 0
        for i in range(1, len(bars)):
            if bars[i].close < bars[i].open and bars[i].vol > bars[i - 1].vol * 2:
                c2 += 1
        if c2 >= t:
            v1 = "空头止盈"

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def pos_stop_V240331(cat: CzscTrader, **kwargs) -> OrderedDict:
    """根据最近N根K线的最高最低价止损，追踪止损，贡献者：谢磊

    参数模板："{pos_name}_{freq1}#{n}_止损V240331"

    **信号逻辑：**

    以多头止损为例，计算过程如下：

    1. 从多头开仓点开始，在给定的K线周期 freq1 上获取最近 N 根K线，记为 bars；
    2. 计算 bars 中的最低价，记为 ll；
    3. 如果当前价格 low < ll，则多头止损。

    空头止损逻辑同理。

    **信号列表：**

    - Signal('SMA5多头_15分钟#10_止损V240331_多头止损_任意_任意_0')
    - Signal('SMA5空头_15分钟#10_止损V240331_空头止损_任意_任意_0')

    :param cat: CzscTrader对象
    :param kwargs: 参数字典

        - pos_name: str，开仓信号的名称
        - freq1: str，给定的K线周期
        - n: int，观察的K线数量，默认为 10，表示观察前10根K线

    :return: OrderedDict
    """
    pos_name = kwargs["pos_name"]
    n = int(kwargs.get("n", 10))
    freq1 = kwargs["freq1"]
    k1, k2, k3 = f"{pos_name}_{freq1}#{n}_止损V240331".split("_")
    v1 = "其他"

    # 如果没有持仓策略，则不产生信号
    if not hasattr(cat, "positions"):
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    # pos_ = [x for x in cat.positions if x.name == pos_name][0]
    pos_ = cat.get_position(pos_name)
    # 如果 pos 没有操作记录，或者最后一次操作是平仓，则不产生信号
    if len(pos_.operates) == 0 or pos_.operates[-1]["op"] in [Operate.SE, Operate.LE]:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    c: CZSC = cat.kas[freq1]
    op = pos_.operates[-1]
    bars = get_sub_elements(c.bars_raw, di=1, n=n + 1)
    _bar = bars[-1]

    # 多头止损逻辑：当前价格低于前n根K线的最低价
    if op["op"] == Operate.LO:
        ll = min([x.low for x in bars[:-1]])
        if _bar.low < ll and _bar.id > op["bid"]:
            v1 = "多头止损"

    # 空头止损逻辑：当前价格高于前n根K线的最高价
    if op["op"] == Operate.SO:
        hh = max([x.high for x in bars[:-1]])
        if _bar.high > hh and _bar.id > op["bid"]:
            v1 = "空头止损"

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def pos_stop_V240608(cat: CzscTrader, **kwargs) -> OrderedDict:
    """止损：多头开仓后，最低价跌破前W根K线最低价N个价位，提示止损；空头反之。

    参数模板："{pos_name}_{freq1}W{w}N{n}_止损V240608"

    **信号逻辑：**

    以多头止损为例，计算过程如下：

    1. 从多头开仓点开始，在给定的K线周期 freq1 上计算开仓前 W 个K线的最低价，记为 L1；
    2. 计算开仓后的最低价，记为 L2；
    3. 如果 L2 < L1 - N 个价位，则提示多头止损信号。

    空头止损逻辑同理。

    **信号列表：**

    - Signal('SMA5多头_15分钟W20N10_止损V240608_多头止损_任意_任意_0')
    - Signal('SMA5空头_15分钟W20N10_止损V240608_空头止损_任意_任意_0')

    :param cat: CzscTrader对象
    :param kwargs: 参数字典

        - pos_name: str，开仓信号的名称
        - freq1: str，给定的K线周期
        - w: int，开仓前W根K线，默认为 20
        - n: int，最低价下方N个价位，默认为 10

    :return: OrderedDict
    """
    pos_name = kwargs["pos_name"]
    freq1 = kwargs["freq1"]
    w = int(kwargs.get("w", 20))  # 开仓前W根K线
    n = int(kwargs.get("n", 10))  # N个价位

    k1, k2, k3 = f"{pos_name}_{freq1}W{w}N{n}_止损V240608".split("_")
    v1 = "其他"

    # 如果没有持仓策略，则不产生信号
    if not cat.kas or not hasattr(cat, "positions"):
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    pos = [x for x in cat.positions if x.name == pos_name][0]
    if len(pos.operates) == 0 or pos.operates[-1]["op"] in [Operate.SE, Operate.LE]:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    c = cat.kas[freq1]
    op = pos.operates[-1]

    # 开仓前W根K线
    w_bars = [x for x in c.bars_raw if x.dt < op["dt"]][-w:]
    # 开仓后的K线
    a_bars = [x for x in c.bars_raw[-100:] if x.dt > op["dt"]]
    unique_prices = [p for x in c.bars_raw[-200:] for p in [x.high, x.low, x.close, x.open]]
    unique_prices = sorted(list(set(unique_prices)))  # 去重并按升序排列

    if op["op"] == Operate.LO and w_bars:
        w_low = min([x.low for x in w_bars])  # 开仓前最低价
        a_low = min([x.low for x in a_bars])  # 开仓后最低价
        w_low_prices = [x for x in unique_prices if x < w_low]  # 开仓前最低价下方的价位，升序排列
        # 如果开仓后的最低价低于开仓前的最低价向下的 N 个价位，则提示多头止损
        if len(w_low_prices) > n and a_low < w_low_prices[-n]:
            v1 = "多头止损"

    if op["op"] == Operate.SO and w_bars:
        w_high = max([x.high for x in w_bars])
        a_high = max([x.high for x in a_bars])
        w_high_prices = [x for x in unique_prices if x > w_high]
        if len(w_high_prices) > n and a_high > w_high_prices[n]:
            v1 = "空头止损"

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def pos_stop_V240614(cat: CzscTrader, **kwargs) -> OrderedDict:
    """止损：多头开仓后，有超过N根K线的最低价在成本价下方，提示止损；空头反之。

    参数模板："{pos_name}_{freq1}N{n}_止损V240614"

    **信号逻辑：**

    以多头止损为例，计算过程如下：

    1. 从多头开仓点开始，在给定的K线周期 freq1 上获取开仓后的所有K线，记为 bars；
    2. 计算 bars 中的最低价小于开仓价的数量，记为 C；
    3. 如果 C >= N，则提示多头止损信号。

    空头止损逻辑同理。

    **信号列表：**

    - Signal('SMA5多头_15分钟N10_止损V240614_多头止损_任意_任意_0')
    - Signal('SMA5空头_15分钟N10_止损V240614_空头止损_任意_任意_0')

    :param cat: CzscTrader对象
    :param kwargs: 参数字典

        - pos_name: str，开仓信号的名称
        - freq1: str，给定的K线周期
        - n: int，最低价下方N个价位，默认为 10

    :return: OrderedDict
    """
    pos_name = kwargs["pos_name"]
    freq1 = kwargs["freq1"]
    n = int(kwargs.get("n", 10))  # N根K线

    k1, k2, k3 = f"{pos_name}_{freq1}N{n}_止损V240614".split("_")
    v1 = "其他"

    # 如果没有持仓策略，则不产生信号
    if not cat.kas or not hasattr(cat, "positions"):
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    pos = [x for x in cat.positions if x.name == pos_name][0]
    if len(pos.operates) == 0 or pos.operates[-1]["op"] in [Operate.SE, Operate.LE]:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    c = cat.kas[freq1]
    op = pos.operates[-1]

    # 开仓后的K线
    a_bars = [x for x in c.bars_raw if x.dt >= op["dt"]]

    if op["op"] == Operate.LO and len([x for x in a_bars if x.low < op["price"]]) >= n:
        v1 = "多头止损"

    if op["op"] == Operate.SO and len([x for x in a_bars if x.high > op["price"]]) >= n:
        v1 = "空头止损"

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)


def pos_stop_V240717(cat: CzscTrader, **kwargs) -> OrderedDict:
    """止损：多头开仓后，有超过N根K线的最低价在成本价-ATR*0.67下方，提示止损；空头反之。贡献者：谢磊

    参数模板："{pos_name}_{freq1}N{n}T{timeperiod}_止损V240717"

    **信号逻辑：**

    以多头止损为例，计算过程如下：

    1. 从多头开仓点开始，在给定的K线周期 freq1 上获取开仓后的所有K线，记为 bars；
    2. 计算 bars 中的最低价小于（开仓价-ATR*0.67）的数量，记为 C；
    3. ATR的参数为默认参数，可以调整；
    3. 如果 C >= N，则提示多头止损信号。

    空头止损逻辑同理。

    **信号列表：**

    - Signal('SMA5多头_15分钟N3T20_止损V240717_多头止损_任意_任意_0')
    - Signal('SMA5空头_15分钟N3T20_止损V240717_空头止损_任意_任意_0')

    :param cat: CzscTrader对象
    :param kwargs: 参数字典

        - pos_name: str，开仓信号的名称
        - freq1: str，给定的K线周期
        - n: int，最低价下方N个价位，默认为 3

    :return: OrderedDict
    """
    from czsc.signals.tas import update_atr_cache

    pos_name = kwargs["pos_name"]
    freq1 = kwargs["freq1"]
    n = int(kwargs.get("n", 10))  # N根K线
    timeperiod = int(kwargs.get("timeperiod", 20))  # ATR参数

    c = cat.kas[freq1]
    cache_key = update_atr_cache(c, timeperiod=timeperiod)

    k1, k2, k3 = f"{pos_name}_{freq1}N{n}T{timeperiod}_止损V240717".split("_")
    v1 = "其他"

    # 如果没有持仓策略，则不产生信号
    if not cat.kas or not hasattr(cat, "positions"):
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    pos = [x for x in cat.positions if x.name == pos_name][0]
    if len(pos.operates) == 0 or pos.operates[-1]["op"] in [Operate.SE, Operate.LE]:
        return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)

    op = pos.operates[-1]
    atr = [x.cache[cache_key] if x.cache.get(cache_key) is not None else 0 for x in c.bars_raw if x.dt == op["dt"]]

    # 开仓后的K线
    a_bars = [x for x in c.bars_raw if x.dt >= op["dt"]]

    if op["op"] == Operate.LO and len([x for x in a_bars if x.low < op["price"] - atr[0] * 0.67]) >= n:
        v1 = "多头止损"

    if op["op"] == Operate.SO and len([x for x in a_bars if x.high > op["price"] + atr[0] * 0.67]) >= n:
        v1 = "空头止损"

    return create_single_signal(k1=k1, k2=k2, k3=k3, v1=v1)



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/2/26 15:06
describe: 测试绘图
"""
import os
import pandas as pd
from czsc import CZSC, KlineChart
from test.test_analyze import read_daily


def test_kline_chart():
    """测试K线图"""
    bars = read_daily()
    c = CZSC(bars, max_bi_num=50)

    df = pd.DataFrame(c.bars_raw)
    df['text'] = "测试"
    kline = KlineChart(n_rows=3)
    kline.add_kline(df, name="K线")
    kline.add_sma(df, ma_seq=(5, 10, 21), row=1, visible=True, line_width=1.2)
    kline.add_sma(df, ma_seq=(34, 55, 89, 144), row=1, visible=False, line_width=1.2)
    kline.add_vol(df, row=2)
    kline.add_macd(df, row=3)
    if len(c.bi_list) > 0:
        bi1 = [{'dt': x.fx_a.dt, "bi": x.fx_a.fx, "text": x.fx_a.mark.value} for x in c.bi_list]
        bi2 = [{'dt': c.bi_list[-1].fx_b.dt, "bi": c.bi_list[-1].fx_b.fx, "text": c.bi_list[-1].fx_b.mark.value}]
        bi = pd.DataFrame(bi1 + bi2)
        fx = pd.DataFrame([{'dt': x.dt, "fx": x.fx} for x in c.fx_list])
        kline.add_scatter_indicator(fx['dt'], fx['fx'], name="分型", row=1, line_width=2)
        kline.add_scatter_indicator(bi['dt'], bi['bi'], name="笔", text=bi['text'], row=1, line_width=2)
    # kline.open_in_browser()
    file_html = "kline_chart_test.html"
    kline.fig.write_html(file_html)
    assert os.path.exists(file_html)
    os.remove(file_html)
    assert not os.path.exists(file_html)



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2022/3/17 21:41
describe: 环境变量统一管理入口
"""

import os

# True 的有效表达
valid_true = ['1', 'True', 'true', 'Y', 'y', 'yes', 'Yes', True]


def get_verbose(verbose=None):
    """verbose - 是否输出执行过程的详细信息"""
    verbose = verbose if verbose else os.environ.get('czsc_verbose', None)
    v = True if verbose in valid_true else False
    return v


def get_welcome():
    """welcome - 是否输出版本标识和缠中说禅博客摘记"""
    v = True if os.environ.get('czsc_welcome', '0') in valid_true else False
    return v


def get_min_bi_len(v: int = None) -> int:
    """min_bi_len - 一笔的最小长度，也就是无包含K线的数量，7是老笔的要求，6是新笔的要求"""
    min_bi_len = v if v else os.environ.get('czsc_min_bi_len', 6)
    return int(float(min_bi_len))


def get_max_bi_num(v: int = None) -> int:
    """max_bi_num - 单个级别K线分析中，程序最大保存的笔数量

    默认值为 50，仅使用内置的信号和因子，不需要调整这个参数。
    如果进行新的信号计算需要用到更多的笔，可以适当调大这个参数。
    """
    max_bi_num = v if v else os.environ.get('czsc_max_bi_num', 50)
    return int(float(max_bi_num))



"""
工具类组件

包含代码编辑器等辅助功能
"""

import streamlit as st


def show_code_editor(text: str = "", **kwargs):
    """展示代码编辑器

    :param text: str, 初始代码文本
    :param kwargs:
        - language: str, 编程语言，默认为 'python'
        - theme: str, 主题，默认为 'monokai'
        - key: str, 组件的唯一标识符
        - height: int, 编辑器高度，默认为 400
        - auto_update: bool, 是否自动更新，默认为 False
        - wrap: bool, 是否自动换行，默认为 False
        - font_size: int, 字体大小，默认为 14
    """
    try:
        from streamlit_ace import st_ace
    except ImportError:
        st.error("请先安装 streamlit-ace: pip install streamlit-ace")
        return None

    language = kwargs.get("language", "python")
    theme = kwargs.get("theme", "monokai")
    key = kwargs.get("key", "code_editor")
    height = kwargs.get("height", 400)
    auto_update = kwargs.get("auto_update", False)
    wrap = kwargs.get("wrap", False)
    font_size = kwargs.get("font_size", 14)

    content = st_ace(
        value=text,
        language=language,
        theme=theme,
        key=key,
        height=height,
        auto_update=auto_update,
        wrap=wrap,
        font_size=font_size,
    )

    return content 